from selenium.webdriver.common.keys import Keys

from cumulusci.robotframework.pageobjects import BasePage, pageobject
from cumulusci.robotframework.utils import capture_screenshot_on_error

object_manager = {
    "button": "//input[@title='{}']",
    "input": "//input[@id='{}']",
    "select_related": "//select[@id = '{}']",
    "select_related_option": "//select[@id = 'DomainEnumOrId']/option[@value='{}']",
    "search_result": "//section[@class='related-list-card']//a[.=\"{}\"]",
    "formula_txtarea": "//textarea[@id = '{}']",
    "object_result": "//th/a[text()='{}']",
    "link-text": "//a[contains(text(),'{}')]",
    "button-with-text": "//button[contains(text(),'{}')]",
    "frame_new": "//iframe[contains(@name, '{}') or contains(@title, '{}')]",
    "action_menu": "//tr[.//a[.='{}']]//div[contains(@class, 'objectManagerVirtualActionMenu')]//a",
    "action_menu_item": "//div[@class='actionMenu']//a[@role='menuitem' and .='{}']",
    "delete_confirm_btn": "//button[contains(@class,'forceActionButton')]",
}
# All common elements
search_button = object_manager["input"].format("globalQuickfind")
currency_locator = object_manager["input"].format("dtypeC")
next_button = object_manager["button"].format("Next")
save_button = object_manager["button"].format("Save")
text_locator = object_manager["input"].format("dtypeS")
formula_locator = object_manager["input"].format("dtypeZ")
checkbox_option = object_manager["input"].format("fdtypeB")
formula_txtarea = object_manager["formula_txtarea"].format("CalculatedFormula")
check_syntax = object_manager["button"].format("Check Syntax")
actions_menu = object_manager["action_menu"]
action_item_delete = object_manager["action_menu_item"].format("Delete")
confirm_delete = object_manager["delete_confirm_btn"]
lookup_locator = object_manager["input"].format("dtypeY")


@pageobject(page_type="ObjectManager")
class ObjectManagerPage(BasePage):
    """A page object representing the Object Manager of an object.
    Example
    | Go to page   ObjectManager  Contact
    """

    def _go_to_page(self):
        url_template = "{root}/lightning/setup/ObjectManager/home"
        url = url_template.format(root=self.cumulusci.org.lightning_base_url)
        self.selenium.go_to(url)
        object_name = self.object_name
        self.salesforce.wait_until_loading_is_complete()
        self.selenium.wait_until_page_contains_element(search_button)
        self.selenium.press_keys(search_button, object_name, "RETURN")
        object_locator = object_manager["object_result"].format(object_name)
        self.selenium.wait_until_element_is_visible(object_locator)
        self.salesforce._jsclick(object_locator)
        self.selenium.wait_until_location_contains("Details/view", timeout=90)

    def _is_current_page(self):
        self.selenium.location_should_contain("Details/view")

    @capture_screenshot_on_error
    def switch_tab_to(self, tab):
        """Clicks on the name of the tab specified and sets the page view to be on the tab specified
        Example
        | Switch Tab To     Fields & Relationships
        """
        leftnavoption = object_manager["link-text"].format(tab)
        self.selenium.click_element(leftnavoption)

    def _is_current_tab(self, tab):
        tab_view = f"{tab}/view"
        self.selenium.location_should_contain(tab_view)

    @capture_screenshot_on_error
    def create_currency_field(self, field_name):
        """Creates a currency field by taking in the field name"""
        self.selenium.wait_until_page_contains_element(currency_locator, timeout=60)
        self.selenium.click_element(currency_locator)
        self.selenium.wait_until_page_contains_element(next_button, 60)
        self.selenium.click_element(next_button)
        self.salesforce.populate_field("Field Label", field_name)
        self.salesforce.populate_field("Length", "16")
        self.salesforce.populate_field("Decimal Places", "2")
        self.salesforce.populate_field(
            "Description", "This is a custom field generated during automation"
        )
        self.selenium.click_element(next_button)
        self.selenium.click_element(next_button)
        self.selenium.click_element(save_button)
        self.selenium.wait_until_location_contains(
            "FieldsAndRelationships/view",
            timeout=90,
            message="Fields And Relationships page did not load in 1 min",
        )

    @capture_screenshot_on_error
    def create_text_field(self, field_name):
        """Creates a text field by taking in the field name"""
        self.selenium.wait_until_page_contains_element(text_locator, timeout=60)
        self.selenium.click_element(text_locator)
        self.selenium.click_element(next_button)
        self.salesforce.populate_field("Field Label", field_name)
        self.salesforce.populate_field("Length", "255")
        self.salesforce.populate_field(
            "Description", "This is a custom field generated during automation"
        )
        self.selenium.click_element(next_button)
        self.selenium.click_element(next_button)
        self.selenium.click_element(save_button)
        self.selenium.wait_until_location_contains(
            "FieldsAndRelationships/view",
            timeout=90,
            message="Fields And Relationships page did not load in 1 min",
        )

    @capture_screenshot_on_error
    def create_formula_field(self, field_name, formula):
        """Creates a formula field by providing the field_name, formula and forumla fields"""
        self.selenium.wait_until_page_contains_element(formula_locator, 60)
        self.selenium.click_element(formula_locator)
        self.selenium.wait_until_page_contains_element(next_button, 60)
        self.selenium.click_element(next_button)
        self.salesforce.populate_field("Field Label", field_name)
        self.selenium.wait_until_page_contains_element(checkbox_option, 60)
        self.selenium.click_element(checkbox_option)
        self.selenium.click_element(next_button)
        self.selenium.wait_until_page_contains_element(formula_txtarea, 60)
        self.selenium.get_webelement(formula_txtarea).send_keys(formula)
        self.selenium.click_element(check_syntax)
        self.selenium.click_element(next_button)
        self.selenium.click_element(next_button)
        self.selenium.click_element(save_button)
        self.selenium.wait_until_location_contains(
            "FieldsAndRelationships/view",
            timeout=90,
            message="Detail page did not load in 1 min",
        )

    def create_lookup_field(self, field_name, related):
        """Creates a Lookup field by taking in the inputs field_name and related field"""
        option = object_manager["select_related_option"].format(related)
        related = object_manager["select_related"].format("DomainEnumOrId")
        self.selenium.wait_until_page_contains_element(lookup_locator, 60)
        self.selenium.click_element(lookup_locator)
        self.selenium.wait_until_page_contains_element(next_button, 60)
        self.selenium.click_element(next_button)
        self.selenium.wait_until_page_contains_element(related, 60)
        self.salesforce.scroll_element_into_view(related)
        self.selenium.get_webelement(related).click()
        self.selenium.click_element(option)
        self.selenium.wait_until_page_contains_element(next_button, 60)
        self.selenium.click_element(next_button)
        self.salesforce.populate_field("Field Label", field_name)
        self.salesforce.populate_field(
            "Description", "This is a custom field generated during automation"
        )
        self.selenium.click_element(next_button)
        self.selenium.click_element(next_button)
        self.selenium.click_element(next_button)
        self.selenium.click_element(save_button)
        self.selenium.wait_until_location_contains(
            "FieldsAndRelationships/view",
            timeout=90,
            message="Detail page did not load in 1 min",
        )

    @capture_screenshot_on_error
    def is_field_present(self, field_name):
        """Searches for the field name (field_name) and asserts the field got created"""
        self.selenium.wait_until_page_contains_element(search_button)
        self.selenium.clear_element_text(search_button)
        self.selenium.press_keys(search_button, field_name, "ENTER")
        self.selenium.wait_until_element_is_not_visible("sf:spinner")
        search_results = object_manager["search_result"].format(field_name)
        self.selenium.wait_until_page_contains_element(search_results)

    @capture_screenshot_on_error
    def delete_custom_field(self, field_name):
        """Searches for the custom field and performs the delete action from the actions menu next to the field"""
        action_menu_button = actions_menu.format(field_name)
        self.is_field_present(field_name)

        self.selenium.wait_until_page_contains_element(action_menu_button)
        self.salesforce.scroll_element_into_view(action_menu_button)

        # I don't know why, but sometimes clicking the action menu just
        # doesn't work. The menu doesn't appear, or clicking the item on
        # the menu does nothing. So, we'll try and handful of times.
        # Yes, this feels icky.
        for tries in range(5):
            try:
                self.selenium.wait_until_element_is_visible(action_menu_button)
                self.selenium.click_element(action_menu_button)
                self.selenium.wait_until_element_is_visible(
                    action_item_delete, timeout="5 seconds"
                )
                self.selenium.click_element(action_item_delete, action_chain=True)
                self.selenium.wait_until_element_is_visible(
                    confirm_delete, timeout="5 seconds"
                )
                self.selenium.click_element(confirm_delete, action_chain=True)
                self.selenium.wait_until_location_contains("/view", timeout=90)
                return

            except Exception as e:
                self.builtin.log(
                    f"on try #{tries+1} we caught this error: {e}", "DEBUG"
                )
                self.builtin.sleep("1 second")
                last_error = e
        raise (last_error)

    @capture_screenshot_on_error
    def create_custom_field(self, **kwargs):
        """creates a custom field if one doesn't already exist by the given name (Field_Name)
        |  Example
        | Create custom field
        | ...  Object=Payment
        | ...  Field_Type=Formula
        | ...  Field_Name=Is Opportunity From Prior Year
        | ...  Formula=YEAR( npe01__Opportunity__r.CloseDate ) < YEAR( npe01__Payment_Date__c )
        """
        self.selenium.wait_until_page_contains_element(search_button, 60)
        self.selenium.get_webelement(search_button).send_keys(kwargs["Field_Name"])
        self.selenium.get_webelement(search_button).send_keys(Keys.ENTER)
        self.salesforce.wait_until_loading_is_complete()
        search_results = object_manager["search_result"].format(kwargs["Field_Name"])
        count = len(self.selenium.get_webelements(search_results))
        if count != 1:
            locator = object_manager["button-with-text"].format("New")
            self.selenium.wait_until_page_contains_element(locator, 60)
            self.selenium.get_webelement(locator).click()
            self.salesforce.wait_until_loading_is_complete()
            framelocator = object_manager["frame_new"].format("vfFrameId", "vfFrameId")
            self.selenium.wait_until_element_is_visible(framelocator, timeout=60)
            frame = self.selenium.get_webelement(framelocator)
            self.selenium.select_frame(frame)
            type = kwargs["Field_Type"]
            if type.lower() == "lookup":
                self.create_lookup_field(kwargs["Field_Name"], kwargs["Related_To"])
            elif type.lower() == "currency":
                self.create_currency_field(kwargs["Field_Name"])
            elif type.lower() == "formula":
                self.create_formula_field(kwargs["Field_Name"], kwargs["Formula"])
            elif type.lower() == "text":
                self.create_text_field(kwargs["Field_Name"])
            self.selenium.unselect_frame()



import typing as T
from functools import cached_property

from sqlalchemy import String, and_, func, text
from sqlalchemy.orm import Query, aliased
from sqlalchemy.sql import literal_column

from cumulusci.core.exceptions import BulkDataException

Criterion = T.Any
ID_TABLE_NAME = "cumulusci_id_table"


class LoadQueryExtender:
    """Class that transforms a load.py query with columns, filters, joins"""

    @cached_property
    def columns_to_add(*args) -> T.Optional[T.List]:
        return None

    @cached_property
    def filters_to_add(*args) -> T.Optional[T.List]:
        return None

    @cached_property
    def outerjoins_to_add(*args) -> T.Optional[T.List]:
        return None

    def __init__(self, mapping, metadata, model) -> None:
        self.mapping, self.metadata, self.model = mapping, metadata, model

    def add_columns(self, query: Query):
        """Add columns to the query"""
        if self.columns_to_add:
            query = query.add_columns(*self.columns_to_add)
        return query

    def add_filters(self, query: Query):
        """Add filters to the query"""
        if self.filters_to_add:
            return query.filter(*self.filters_to_add)
        return query

    def add_outerjoins(self, query: Query):
        """Add outer joins to the query"""
        if self.outerjoins_to_add:
            for table, condition in self.outerjoins_to_add:
                query = query.outerjoin(table, condition)
        return query


class AddLookupsToQuery(LoadQueryExtender):
    """Adds columns and joins relatinng to lookups"""

    def __init__(self, mapping, metadata, model, _old_format) -> None:
        super().__init__(mapping, metadata, model)
        self._old_format = _old_format
        self.lookups = [
            lookup for lookup in self.mapping.lookups.values() if not lookup.after
        ]

    @cached_property
    def columns_to_add(self):
        for lookup in self.lookups:
            lookup.aliased_table = aliased(self.metadata.tables[ID_TABLE_NAME])
        return [lookup.aliased_table.columns.sf_id for lookup in self.lookups]

    @cached_property
    def outerjoins_to_add(self):
        # Outer join with lookup ids table:
        # returns main obj even if lookup is null
        def join_for_lookup(lookup):
            key_field = lookup.get_lookup_key_field(self.model)
            value_column = getattr(self.model, key_field)
            if self._old_format:
                return (
                    lookup.aliased_table,
                    lookup.aliased_table.columns.id
                    == str(lookup.table) + "-" + func.cast(value_column, String),
                )
            else:
                return (
                    lookup.aliased_table,
                    lookup.aliased_table.columns.id == value_column,
                )

        return [join_for_lookup(lookup) for lookup in self.lookups]


class DynamicLookupQueryExtender(LoadQueryExtender):
    """Dynamically adds columns and joins for all fields in lookup tables, handling polymorphic lookups"""

    def __init__(
        self, mapping, all_mappings, metadata, model, _old_format: bool
    ) -> None:
        super().__init__(mapping, metadata, model)
        self._old_format = _old_format
        self.all_mappings = all_mappings
        self.lookups = [
            lookup for lookup in self.mapping.lookups.values() if not lookup.after
        ]

    @cached_property
    def columns_to_add(self):
        """Add all relevant fields from lookup tables directly without CASE, with support for polymorphic lookups."""
        columns = []
        for lookup in self.lookups:
            tables = lookup.table if isinstance(lookup.table, list) else [lookup.table]
            lookup.parent_tables = [
                aliased(
                    self.metadata.tables[table], name=f"{lookup.name}_{table}_alias"
                )
                for table in tables
            ]

            for parent_table, table_name in zip(lookup.parent_tables, tables):
                # Find the mapping step for this polymorphic type
                lookup_mapping_step = next(
                    (
                        step
                        for step in self.all_mappings.values()
                        if step.table == table_name
                    ),
                    None,
                )
                if lookup_mapping_step:
                    load_fields = lookup_mapping_step.fields.keys()
                    for field in load_fields:
                        if field in lookup_mapping_step.fields:
                            matching_column = next(
                                (
                                    col
                                    for col in parent_table.columns
                                    if col.name == lookup_mapping_step.fields[field]
                                )
                            )
                            columns.append(
                                matching_column.label(f"{parent_table.name}_{field}")
                            )
                        else:
                            # Append an empty string if the field is not present
                            columns.append(
                                literal_column("''").label(
                                    f"{parent_table.name}_{field}"
                                )
                            )
        return columns

    @cached_property
    def outerjoins_to_add(self):
        """Add outer joins for each lookup table directly, including handling for polymorphic lookups."""

        def join_for_lookup(lookup, parent_table):
            key_field = lookup.get_lookup_key_field(self.model)
            value_column = getattr(self.model, key_field)
            return (parent_table, parent_table.columns.id == value_column)

        joins = []
        for lookup in self.lookups:
            for parent_table in lookup.parent_tables:
                joins.append(join_for_lookup(lookup, parent_table))
        return joins


class AddRecordTypesToQuery(LoadQueryExtender):
    """Adds columns, joins and filters relatinng to recordtypes"""

    def __init__(self, mapping, metadata, model) -> None:
        super().__init__(mapping, metadata, model)
        if "RecordTypeId" in mapping.fields:
            self.rt_dest_table = metadata.tables[
                mapping.get_destination_record_type_table()
            ]
        else:
            self.rt_dest_table = None

    @cached_property
    def columns_to_add(self):
        if self.rt_dest_table is not None:
            return [self.rt_dest_table.columns.record_type_id]

    @cached_property
    def filters_to_add(self):
        if self.mapping.record_type and hasattr(self.model, "record_type"):
            return [self.model.record_type == self.mapping.record_type]

    @cached_property
    def outerjoins_to_add(self):

        if "RecordTypeId" in self.mapping.fields:
            try:
                rt_source_table = self.metadata.tables[
                    self.mapping.get_source_record_type_table()
                ]

            except KeyError as e:
                # For generate_and_load_from_yaml, In case of namespace_inject true, mapping table name doesn't have namespace added
                # We are checking for table_rt_mapping table
                try:
                    rt_source_table = self.metadata.tables[
                        f"{self.mapping.table}_rt_mapping"
                    ]

                except KeyError as f:

                    raise BulkDataException(
                        "A record type mapping table was not found in your dataset. "
                        f"Was it generated by extract_data? {e}",
                    ) from f

            rt_dest_table = self.metadata.tables[
                self.mapping.get_destination_record_type_table()
            ]

            # Check if 'is_person_type' column exists in rt_source_table.columns
            is_person_type_column = getattr(
                rt_source_table.columns, "is_person_type", None
            )
            # If it does not exist, set condition to True
            is_person_type_condition = (
                rt_dest_table.columns.is_person_type == is_person_type_column
                if is_person_type_column is not None
                else True
            )

            return [
                (
                    rt_source_table,
                    rt_source_table.columns.record_type_id
                    == getattr(self.model, self.mapping.fields["RecordTypeId"]),
                ),
                # Combination of IsPersonType and DeveloperName is unique
                (
                    rt_dest_table,
                    and_(
                        rt_dest_table.columns.developer_name
                        == rt_source_table.columns.developer_name,
                        is_person_type_condition,
                    ),
                ),
            ]


class AddMappingFiltersToQuery(LoadQueryExtender):
    """Adds filters relating to user-specified filters"""

    @cached_property
    def filters_to_add(self):
        if self.mapping.filters:
            return [text(f) for f in self.mapping.filters]


class AddPersonAccountsToQuery(LoadQueryExtender):
    """Add filters relating to Person accounts."""

    @cached_property
    def filters_to_add(self):
        """Filter out non-person account Contact records.
        Contact records for person accounts were already created by the system."""

        assert self.mapping.sf_object == "Contact"
        return [
            func.lower(self.model.__table__.columns.get("IsPersonAccount")) == "false"
        ]



import re
from unittest import mock

import pytest
import responses
from simple_salesforce.exceptions import SalesforceError

from cumulusci.core.exceptions import SalesforceException
from cumulusci.tasks.salesforce.insert_record import InsertRecord
from cumulusci.tests.util import CURRENT_SF_API_VERSION

from .util import create_task


class TestCreateRecord:
    def test_run_task(self):
        task = create_task(
            InsertRecord,
            {
                "object": "PermissionSet",
                "values": "Name:HardDelete,PermissionsBulkApiHardDelete:true",
            },
        )
        create = mock.Mock()
        task.sf = mock.Mock(create=create)
        task.sf.PermissionSet.create.return_value = {
            "id": "0PS3D000000MKTqWAO",
            "success": True,
            "errors": [],
        }

        task._run_task()
        task.sf.PermissionSet.create.assert_called_with(
            {"Name": "HardDelete", "PermissionsBulkApiHardDelete": "true"}
        )

    def test_run_task__dict_tooling(self):
        task = create_task(
            InsertRecord,
            {
                "object": "PermissionSet",
                "tooling": True,
                "values": {"Name": "HardDelete", "PermissionsBulkApiHardDelete": True},
            },
        )
        create = mock.Mock()
        task.tooling = mock.Mock(create=create)
        task.tooling.PermissionSet.create.return_value = {
            "id": "0PS3D000000MKTqWAO",
            "success": True,
            "errors": [],
        }

        task._run_task()
        task.tooling.PermissionSet.create.assert_called_with(
            {"Name": "HardDelete", "PermissionsBulkApiHardDelete": True}
        )

    def test_salesforce_error_returned_by_simple_salesforce(self):
        "Tests the just-in-case path where SimpleSalesforce does not raise an exception"
        task = create_task(
            InsertRecord,
            {
                "object": "PermissionSet",
                "values": "Name:HardDelete,PermissionsBulkApiHardDelete:true",
            },
        )
        create = mock.Mock()
        task.sf = mock.Mock(create=create)
        task.sf.PermissionSet.create.return_value = {
            "success": False,
            "errors": [
                {
                    "errorCode": "NOT_FOUND",
                    "message": "The requested resource does not exist",
                }
            ],
        }
        with pytest.raises(SalesforceException):
            task._run_task()

    @responses.activate
    def test_salesforce_error_raised_by_simple_salesforce(self):
        task = create_task(
            InsertRecord,
            {
                "object": "PermissionSet",
                "values": "Name:HardDelete,PermissionsBulkApiHardDelete:true",
            },
        )
        responses.add(
            responses.POST,
            re.compile(
                rf"https://test.salesforce.com/services/data/v{CURRENT_SF_API_VERSION}/.*"
            ),
            content_type="application/json",
            status=404,
            json={
                "success": False,
                "errors": [
                    {
                        "errorCode": "NOT_FOUND",
                        "message": "The requested resource does not exist",
                    }
                ],
            },
        )
        task._init_task()
        with pytest.raises(SalesforceError):
            task._run_task()



import pytest

from cumulusci.tasks.bulkdata import LoadData


class TestSelect:
    @pytest.mark.vcr()
    def test_select_similarity_strategy(
        self, create_task, cumulusci_test_repo_root, sf
    ):
        self._test_select_similarity_strategy(
            "rest", create_task, cumulusci_test_repo_root, sf
        )

    @pytest.mark.vcr()
    def test_select_similarity_select_and_insert_strategy(
        self, create_task, cumulusci_test_repo_root, sf
    ):
        self._test_select_similarity_select_and_insert_strategy(
            "rest", create_task, cumulusci_test_repo_root, sf
        )

    @pytest.mark.vcr(allow_playback_repeats=True)
    def test_select_similarity_select_and_insert_strategy_bulk(
        self, create_task, cumulusci_test_repo_root, sf
    ):
        self._test_select_similarity_select_and_insert_strategy_bulk(
            "bulk", create_task, cumulusci_test_repo_root, sf
        )

    @pytest.mark.vcr()
    def test_select_random_strategy(self, create_task, cumulusci_test_repo_root, sf):
        self._test_select_random_strategy(
            "rest", create_task, cumulusci_test_repo_root, sf
        )

    @pytest.mark.vcr()
    def test_select_standard_strategy(self, create_task, cumulusci_test_repo_root, sf):
        self._test_select_standard_strategy(
            "rest", create_task, cumulusci_test_repo_root, sf
        )

    def _test_select_similarity_strategy(
        self, api, create_task, cumulusci_test_repo_root, sf
    ):
        # seed sample data, using a mixture of inserts and
        # upserts-into-empty (which should behave as inserts)
        task = create_task(
            LoadData,
            {
                "sql_path": cumulusci_test_repo_root
                / "datasets/select/similarity_sample.sql",
                "mapping": cumulusci_test_repo_root
                / "datasets/select/similarity_mapping.yml",
                "set_recently_viewed": False,
                "enable_rollback": False,
            },
        )

        task()

        result = task.return_values

        assert (
            str(result)
            == "{'step_results': {'Account': {'sobject': 'Account', 'record_type': None, 'status': <DataOperationStatus.SUCCESS: 'Success'>, 'job_errors': [], 'records_processed': 5, 'total_row_errors': 0}, 'Contact': {'sobject': 'Contact', 'record_type': None, 'status': <DataOperationStatus.SUCCESS: 'Success'>, 'job_errors': [], 'records_processed': 3, 'total_row_errors': 0}}}"
        )

    def _test_select_similarity_select_and_insert_strategy(
        self, api, create_task, cumulusci_test_repo_root, sf
    ):
        # seed sample data, using a mixture of inserts and
        # upserts-into-empty (which should behave as inserts)
        task = create_task(
            LoadData,
            {
                "sql_path": cumulusci_test_repo_root
                / "datasets/select/similarity_select_insert_sample.sql",
                "mapping": cumulusci_test_repo_root
                / "datasets/select/similarity_select_insert_mapping.yml",
                "set_recently_viewed": False,
                "enable_rollback": False,
            },
        )

        task()

        result = task.return_values

        assert (
            str(result)
            == "{'step_results': {'Account': {'sobject': 'Account', 'record_type': None, 'status': <DataOperationStatus.SUCCESS: 'Success'>, 'job_errors': [], 'records_processed': 1, 'total_row_errors': 0}, 'Contact': {'sobject': 'Contact', 'record_type': None, 'status': <DataOperationStatus.SUCCESS: 'Success'>, 'job_errors': [], 'records_processed': 3, 'total_row_errors': 0}, 'Lead': {'sobject': 'Lead', 'record_type': None, 'status': <DataOperationStatus.SUCCESS: 'Success'>, 'job_errors': [], 'records_processed': 2, 'total_row_errors': 0}, 'Event': {'sobject': 'Event', 'record_type': None, 'status': <DataOperationStatus.SUCCESS: 'Success'>, 'job_errors': [], 'records_processed': 3, 'total_row_errors': 0}}}"
        )

    def _test_select_similarity_select_and_insert_strategy_bulk(
        self, api, create_task, cumulusci_test_repo_root, sf
    ):
        # seed sample data, using a mixture of inserts and
        # upserts-into-empty (which should behave as inserts)
        task = create_task(
            LoadData,
            {
                "sql_path": cumulusci_test_repo_root
                / "datasets/select/similarity_select_insert_sample.sql",
                "mapping": cumulusci_test_repo_root
                / "datasets/select/similarity_select_insert_mapping.yml",
                "set_recently_viewed": False,
                "enable_rollback": False,
            },
        )

        task()

        result = task.return_values

        assert (
            str(result)
            == "{'step_results': {'Account': {'sobject': 'Account', 'record_type': None, 'status': <DataOperationStatus.SUCCESS: 'Success'>, 'job_errors': [], 'records_processed': 1, 'total_row_errors': 0}, 'Contact': {'sobject': 'Contact', 'record_type': None, 'status': <DataOperationStatus.SUCCESS: 'Success'>, 'job_errors': [], 'records_processed': 3, 'total_row_errors': 0}, 'Lead': {'sobject': 'Lead', 'record_type': None, 'status': <DataOperationStatus.SUCCESS: 'Success'>, 'job_errors': [], 'records_processed': 2, 'total_row_errors': 0}, 'Event': {'sobject': 'Event', 'record_type': None, 'status': <DataOperationStatus.SUCCESS: 'Success'>, 'job_errors': [], 'records_processed': 3, 'total_row_errors': 0}}}"
        )

    def _test_select_random_strategy(
        self, api, create_task, cumulusci_test_repo_root, sf
    ):
        # seed sample data, using a mixture of inserts and
        # upserts-into-empty (which should behave as inserts)
        task = create_task(
            LoadData,
            {
                "sql_path": cumulusci_test_repo_root
                / "datasets/select/random_sample.sql",
                "mapping": cumulusci_test_repo_root
                / "datasets/select/random_mapping.yml",
                "set_recently_viewed": False,
                "enable_rollback": False,
            },
        )

        task()

        result = task.return_values

        assert (
            str(result)
            == "{'step_results': {'Account': {'sobject': 'Account', 'record_type': None, 'status': <DataOperationStatus.SUCCESS: 'Success'>, 'job_errors': [], 'records_processed': 5, 'total_row_errors': 0}, 'Contact': {'sobject': 'Contact', 'record_type': None, 'status': <DataOperationStatus.SUCCESS: 'Success'>, 'job_errors': [], 'records_processed': 3, 'total_row_errors': 0}}}"
        )

    def _test_select_standard_strategy(
        self, api, create_task, cumulusci_test_repo_root, sf
    ):
        # seed sample data, using a mixture of inserts and
        # upserts-into-empty (which should behave as inserts)
        task = create_task(
            LoadData,
            {
                "sql_path": cumulusci_test_repo_root
                / "datasets/select/random_sample.sql",
                "mapping": cumulusci_test_repo_root
                / "datasets/select/standard_mapping.yml",
                "set_recently_viewed": False,
                "enable_rollback": False,
            },
        )

        task()

        result = task.return_values

        assert (
            str(result)
            == "{'step_results': {'Account': {'sobject': 'Account', 'record_type': None, 'status': <DataOperationStatus.SUCCESS: 'Success'>, 'job_errors': [], 'records_processed': 5, 'total_row_errors': 0}, 'Contact': {'sobject': 'Contact', 'record_type': None, 'status': <DataOperationStatus.SUCCESS: 'Success'>, 'job_errors': [], 'records_processed': 3, 'total_row_errors': 0}}}"
        )



""" CLI logger """
import logging
import os
import sys
import tempfile

import requests
from rich.logging import RichHandler

try:
    import colorama
except ImportError:
    # coloredlogs only installs colorama on Windows
    pass


def init_logger(debug=False):
    """Initialize the logger"""

    logger = logging.getLogger(__name__.split(".")[0])
    for handler in logger.handlers:  # pragma: no cover
        logger.removeHandler(handler)

    if os.name == "nt" and "colorama" in sys.modules:  # pragma: no cover
        colorama.init()

    logger.addHandler(
        RichHandler(
            rich_tracebacks=True,
            show_level=debug,
            show_path=debug,
            tracebacks_show_locals=debug,
        )
    )
    logger.setLevel(logging.DEBUG)
    logger.propagate = False

    if debug:  # pragma: no cover
        # Referenced from:
        # https://github.com/urllib3/urllib3/blob/cd55f2fe98df4d499ab5c826433ee4995d3f6a60/src/urllib3/__init__.py#L48
        def add_rich_logger(
            module: str, level: int = logging.DEBUG
        ) -> logging.StreamHandler:
            """Retrieve the logger for the given module.
            Remove all handlers from it, and add a single RichHandler."""
            logger = logging.getLogger(module)
            for handler in logger.handlers:
                logger.removeHandler(handler)

            handler = RichHandler()
            logger.addHandler(handler)
            logger.setLevel(level)
            logger.debug(f"Added rich.logging.RichHandler to logger: {module}")
            return handler

        # monkey patch urllib3 logger
        requests.packages.urllib3.add_stderr_logger = add_rich_logger
        requests.packages.urllib3.add_stderr_logger("urllib3")


def get_tempfile_logger():
    """Creates a logger that writes to a temporary
    logfile. Returns the logger and path to tempfile"""
    logger = logging.getLogger("tempfile_logger")
    file_handle, filepath = tempfile.mkstemp()
    # close the file as it will be opened again by FileHandler
    os.close(file_handle)
    handler = logging.FileHandler(filepath, encoding="utf-8")
    handler.terminator = ""
    logger.addHandler(handler)
    logger.setLevel(logging.DEBUG)
    return logger, filepath



import typing as T
from collections import defaultdict

from cumulusci.tasks.bulkdata.extract_dataset_utils.calculate_dependencies import (
    SObjDependency,
)
from cumulusci.utils.collections import OrderedSet, OrderedSetType

StrDependencyMapping = T.Mapping[str, OrderedSetType[SObjDependency]]
SObjectRuleDeclaration = (
    "snowfakery.cci_mapping_files.declaration_parser.SObjectRuleDeclaration"
)


class DependencyMap:
    """Index which tables depend on which other ones (through lookups)

    To make lookups easy later.
    """

    # a dictionary allowing easy lookup of inferred/soft dependencies by parent table
    soft_dependencies: StrDependencyMapping

    # a dictionary allowing lookups by (tablename, fieldname) pairs
    reference_fields: StrDependencyMapping

    _sorted_tables = None  # cache for sorted table computation

    def __init__(
        self,
        table_names: T.Iterable[str],
        dependencies: T.Iterable[SObjDependency],
    ):
        self.dependencies = defaultdict(OrderedSet)
        self.reference_fields = {}
        self._map_references(dependencies)
        self.table_names = tuple(table_names)

    def _map_references(
        self,
        intertable_dependencies: T.Iterable[SObjDependency],
    ):
        for dep in intertable_dependencies:
            table_deps = self.dependencies[dep.table_name_from]
            table_deps.add(dep)
            self.reference_fields[
                (dep.table_name_from, dep.field_name)
            ] = dep.table_names_to

    def target_table_for(
        self, tablename: str, fieldname: str
    ) -> T.Optional[T.Union[str, T.Tuple[str, ...]]]:
        return self.reference_fields.get((tablename, fieldname))

    def get_dependency_order(self):
        """Sort the dependencies to output tables in the right order."""
        if not self._sorted_tables:
            # SnowfakeryPersonAccounts: Add this back in when Snowfakery is integrated with this code.
            # _remove_person_contact_id(self.dependencies)
            self._sorted_tables = _sort_by_dependencies(
                tuple(self.table_names), self.dependencies
            )
        return self._sorted_tables


def _sort_by_dependencies(
    table_names: T.Sequence[str],
    dependencies: T.Mapping[str, OrderedSet],
    priority: int = 0,
):
    "Sort tables by dependency relationships."
    sorted_tables = []
    dependencies = dict(dependencies)

    while table_names:
        remaining = len(table_names)
        leaf_tables = [
            table
            for table in table_names
            if _table_is_free(table, dependencies, sorted_tables, priority)
        ]
        sorted_tables.extend(leaf_tables)
        table_names = [table for table in table_names if table not in sorted_tables]

        # Nothing changed, so we need to attempt other techniques.
        if len(table_names) == remaining:
            # this is a bit tricky.
            # run the algorithm with ONLY the declared/hard
            # dependencies and see if it comes to resolution

            higher_priority_sort = _sort_dependencies_higher_priority(
                table_names, dependencies, priority
            )

            if higher_priority_sort:
                sorted_tables.extend(higher_priority_sort)
            else:
                # I'm stuck. Try randomly
                sorted_tables.append(sorted(table_names)[0])

    return sorted_tables


def _sort_dependencies_higher_priority(
    table_names: T.Iterable[str],
    dependencies: T.Mapping[str, OrderedSet],
    priority: int,
):
    """After a priority-ignoring dependency sort has failed, try sorting ONLY
    on that subset of dependencies marked with a higher priority. This function
    is mututally recursive with `_sort_by_dependencies` so it will ratchet
    up the priority potentially multiple times."""
    relevant_dependencies = (
        dep.priority
        for depset in dependencies.values()
        for dep in depset
        if dep.priority > priority
    )
    lowest_priority_dependency = min(relevant_dependencies, default=None)

    if lowest_priority_dependency is not None:
        return _sort_by_dependencies(
            table_names, dependencies, lowest_priority_dependency
        )


def _table_is_free(
    table_name: str,
    dependencies: StrDependencyMapping,
    sorted_tables: T.Sequence[str],
    priority: int,
):
    """Check if every child of this table is already sorted

    Look at the unit test test_table_is_free_simple to see some
    usage examples.
    """
    tables_this_table_depends_upon = dependencies.get(table_name, OrderedSet()).copy()
    for dependency in sorted(tables_this_table_depends_upon):
        table_names = (
            [dependency.table_names_to]
            if isinstance(dependency.table_names_to, str)
            else dependency.table_names_to
        )
        if (
            all(table in sorted_tables or table == table_name for table in table_names)
            or dependency.priority < priority
        ):
            tables_this_table_depends_upon.remove(dependency)

    return len(tables_this_table_depends_upon) == 0


# SnowfakeryPersonAccounts: This code will be enabled when Snowfakery
# is integrated with this code.
# def _remove_person_contact_id(dependencies: T.Mapping[str, SObjDependency]):
#     """Don't allow relationships between a personcontact and an account to mess
#     up the load order.

#     Code adapted from Snowfakery
#     """
#     if "Account" in dependencies:
#         dep_to_person_contact = [
#             dep
#             for dep in dependencies["Account"]
#             if dep.table_name_to.lower() == "personcontact"
#         ]
#         for dep in dep_to_person_contact:
#             dependencies["Account"].remove(dep)



import typing as T

import pydantic

from cumulusci.core.exceptions import TaskOptionsError
from cumulusci.tasks.metadata_etl import MetadataSingleEntityTransformTask
from cumulusci.utils.xml.metadata_tree import MetadataElement


class IPRange(pydantic.BaseModel):
    description: T.Optional[str]
    start_address: pydantic.IPvAnyAddress
    end_address: pydantic.IPvAnyAddress


class IPNetwork(pydantic.BaseModel):
    description: T.Optional[str]
    network: pydantic.IPvAnyNetwork


class AddIPRangesOptions(pydantic.BaseModel):
    ranges: T.List[T.Union[IPRange, IPNetwork]]
    replace: bool = False


class AddIPRanges(MetadataSingleEntityTransformTask):
    entity = "Profile"
    task_options = {
        "ranges": {
            "description": "A list of IP ranges, specified as dicts with the keys 'description' (optional) and either 'start_address' and 'end_address' or 'network' (in CIDR notation).",
            "required": True,
        },
        "replace": {
            "description": "If True, replace all existing ranges. Otherwise, just add ranges. Default is False."
        },
    }

    def _transform_entity(
        self, metadata: MetadataElement, api_name: str
    ) -> MetadataElement:
        try:
            opts = AddIPRangesOptions.parse_obj(self.options)
        except pydantic.ValidationError as exc:
            raise TaskOptionsError(f"Invalid options: {exc}")

        if opts.replace:
            # Remove existing ranges
            for elem in metadata.findall("loginIpRanges"):
                metadata.remove(elem)

        for ip_range in opts.ranges:
            if isinstance(ip_range, IPNetwork):
                elem = metadata.append("loginIpRanges")
                elem.append("description", text=ip_range.description or "")
                elem.append(
                    "startAddress", text=str(ip_range.network.network_address + 1)
                )
                elem.append(
                    "endAddress", text=str(ip_range.network.broadcast_address - 1)
                )
            else:
                elem = metadata.append("loginIpRanges")
                elem.append("description", text=ip_range.description or "")
                elem.append("startAddress", text=str(ip_range.start_address))
                elem.append("endAddress", text=str(ip_range.end_address))

        return metadata


class AddPermissionSetPermissions(MetadataSingleEntityTransformTask):
    entity = "PermissionSet"
    task_options = {
        "field_permissions": {
            "description": "Array of fieldPermissions objects to upsert into permission_set.  Each fieldPermission requires the following attributes: 'field': API Name of the field including namespace; 'readable': boolean if field can be read; 'editable': boolean if field can be edited",
            "required": False,
        },
        "class_accesses": {
            "description": "Array of classAccesses objects to upsert into permission_set.  Each classAccess requires the following attributes: 'apexClass': Name of Apex Class.  If namespaced, make sure to use the form \"namespace__ApexClass\"; 'enabled': boolean if the Apex Class can be accessed.",
            "required": False,
        },
        **MetadataSingleEntityTransformTask.task_options,
    }

    def _transform_entity(
        self, metadata: MetadataElement, api_name: str
    ) -> MetadataElement:
        self._upsert_class_accesses(metadata, api_name)
        self._upsert_field_permissions(metadata, api_name)
        return metadata

    def _upsert_class_accesses(self, metadata, api_name):
        class_accesses = self.options.get("class_accesses")

        if not class_accesses:
            return

        self.logger.info(f"Upserting class accesses for {api_name}")

        for class_access in class_accesses:
            if "apexClass" not in class_access:
                raise TaskOptionsError(
                    "class_access entries must contain the 'apexClass' key."
                )

            class_access["apexClass"] = self._inject_namespace(
                class_access["apexClass"]
            )

            existing_permissions = metadata.findall(
                "classAccesses", apexClass=class_access["apexClass"]
            )

            if len(existing_permissions):
                # Permission exists: update
                for elem in existing_permissions:
                    elem.find("enabled").text = str(
                        class_access.get("enabled", True)
                    ).lower()
            else:
                # Permission doesn't exist: insert
                elem = metadata.append("classAccesses")
                elem.append("apexClass", text=class_access.get("apexClass"))
                elem.append(
                    "enabled", text=str(class_access.get("enabled", True)).lower()
                )

    def _upsert_field_permissions(self, metadata, api_name):
        field_permissions = self.options.get("field_permissions")

        if not field_permissions:
            return

        self.logger.info(f"Upserting Field Level Security for {api_name}")

        for field_permission in field_permissions:
            if "field" not in field_permission:
                raise TaskOptionsError(
                    "field_permissions entries must include the 'field' key."
                )

            field_permission["field"] = self._inject_namespace(
                field_permission["field"]
            )

            existing_permissions = metadata.findall(
                "fieldPermissions", field=field_permission["field"]
            )

            if len(existing_permissions):
                # Permission exists: update
                for elem in existing_permissions:
                    elem.find("readable").text = str(
                        field_permission.get("readable", True)
                    ).lower()
                    elem.find("editable").text = str(
                        field_permission.get("editable", True)
                    ).lower()
            else:
                # Permission doesn't exist: insert
                element = metadata.append("fieldPermissions")
                element.append("field", text=field_permission.get("field"))
                element.append(
                    "editable", text=str(field_permission.get("editable", True)).lower()
                )
                element.append(
                    "readable", text=str(field_permission.get("readable", True)).lower()
                )



from concurrent.futures import Future, ThreadPoolExecutor
from unittest.mock import MagicMock, patch

import pytest

from cumulusci.utils.parallel.queries_in_parallel.run_queries_in_parallel import (
    RunParallelQueries,
)


def run_query(query):
    return {"test": query}


def test_run_queries_in_parallel():
    queries = {
        "query1": "SELECT * FROM Table1",
        "query2": "SELECT * FROM Table2",
    }

    expected_results = {
        "query1": [{"field1": "value1"}, {"field1": "value2"}],
        "query2": [{"field2": "value3"}, {"field2": "value4"}],
    }

    mock_future1 = MagicMock()
    mock_future1.result.return_value = {"records": expected_results["query1"]}
    mock_future2 = MagicMock()
    mock_future2.result.return_value = {"records": expected_results["query2"]}

    with patch.object(ThreadPoolExecutor, "submit") as mock_submit:
        mock_submit.side_effect = [mock_future1, mock_future2]

        results_dict = RunParallelQueries._run_queries_in_parallel(queries, run_query)

    assert results_dict == expected_results


def test_run_queries_in_parallel_with_exception():
    queries = {
        "Query1": "SELECT Id FROM Table1",
    }

    with patch.object(Future, "result", side_effect=Exception("Test exception")):
        with pytest.raises(Exception) as excinfo:
            RunParallelQueries._run_queries_in_parallel(queries, run_query=run_query)
        assert (
            "Error executing query 'Query1': <class 'Exception'>: Test exception"
            == str(excinfo.value)
        )



import io
from unittest import mock

from cumulusci.tasks.robotframework import debugger
from cumulusci.tasks.robotframework.debugger.ui import (
    BrowserProxy,
    SeleniumProxy,
    initialize_highlight_js,
)

"""
These tests verify that the proxy used by the debugger uses
the appropriate javascript methods for the given browser library
"""


class StringContaining(str):
    """Used for fuzzy assertions"""

    def __eq__(self, other):
        return self in other


class TestSeleniumProxy:
    @classmethod
    def setup_class(cls):
        cls.selenium = mock.Mock()
        cls.proxy = SeleniumProxy(cls.selenium)

    def setup_method(self):
        self.mock_listener = mock.Mock()
        self.mock_builtin = mock.Mock()
        self.stdout = io.StringIO()
        self.cli = debugger.DebuggerCli(self.mock_listener, stdout=self.stdout)
        self.cli.builtin = self.mock_builtin
        self.selenium.reset_mock()

    def test_webbrowser_browser(self):
        """Verify the cli has initialized the webbrowser attribute to SeleniumProxy

        ... but only Browser has been loaded and SeleniumLibrary has not
        """
        self.cli.builtin.get_library_instance = mock.Mock(
            return_value={
                "SeleniumLibrary": mock.Mock(),
                "Browser": mock.Mock(),
            }
        )
        assert isinstance(self.cli.webbrowser, SeleniumProxy)

    def test_highlight_injected_css(self):
        """Verify we attempt to inject css into the DOM when highlighting an element"""

        self.selenium.get_webelements = mock.Mock(return_value=[])
        self.proxy.highlight_elements("button")

        self.selenium.driver.execute_script.assert_any_call(initialize_highlight_js)

    def test_highlight_element(self):
        """Verify we attempt to add the custom css to matching elements"""

        element1 = mock.Mock()
        element2 = mock.Mock()
        self.selenium.get_webelements = mock.Mock(return_value=[element1, element2])

        self.proxy.highlight_elements("button")
        # expect 3 calls to execute javascript: once for the initialization
        # and one for each element we find
        assert len(self.selenium.driver.execute_script.mock_calls) == 3
        self.selenium.driver.execute_script.assert_any_call(
            "arguments[0].classList.add('rdbHighlight');", element1
        )
        self.selenium.driver.execute_script.assert_any_call(
            "arguments[0].classList.add('rdbHighlight');", element2
        )


class TestBrowserProxy:
    @classmethod
    def setup_class(cls):
        cls.browser = mock.Mock()
        cls.proxy = BrowserProxy(cls.browser)

    def setup_method(self):
        self.mock_listener = mock.Mock()
        self.mock_builtin = mock.Mock()
        self.stdout = io.StringIO()
        self.cli = debugger.DebuggerCli(self.mock_listener, stdout=self.stdout)
        self.cli.builtin = self.mock_builtin
        self.browser.reset_mock()

    def test_webbrowser_browser(self):
        """Verify the cli has initialized the webbrowser attribute to BrowserProxy

        ... but only Browser has been loaded and SeleniumLibrary has not
        """
        self.cli.builtin.get_library_instance = mock.Mock(
            return_value={
                "Browser": mock.Mock(),
            }
        )
        assert isinstance(self.cli.webbrowser, BrowserProxy)

    def test_highlight_injected_css(self):
        """Verify we attempt to inject css into the DOM when highlighting an element"""
        script = f"() => {{{initialize_highlight_js}}};"
        self.browser.get_elements = mock.Mock(return_value=[])
        self.proxy.highlight_elements("button")
        self.browser.evaluate_javascript.assert_any_call(None, script)

    def test_highlight_element(self):
        """Verify that we call evaluate some javascript that applies the class to every matching element"""
        element1 = mock.Mock()
        element2 = mock.Mock()
        self.browser.get_elements = mock.Mock(return_value=[element1, element2])

        self.proxy.highlight_elements("button")
        script = """(elements) => {
                    for (element of elements) {
                        element.classList.add('rdbHighlight')
                    }
                }"""

        self.browser.evaluate_javascript.assert_any_call(
            "button", script, all_elements=True
        )

    def test_restore_element_style(self):
        self.proxy.restore_element_style()
        script = """(elements) => {
                for (element of elements) {
                    element.classList.remove('rdbHighlight')
                }
            }"""
        self.browser.evaluate_javascript.assert_called_with(
            ".rdbHighlight", script, all_elements=True
        )



import os


class BaseChangeNotesProvider(object):
    def __init__(self, release_notes_generator):
        self.release_notes_generator = release_notes_generator

    def __call__(self):
        """Subclasses should provide an implementation that returns an
        iterable of each change note"""
        raise NotImplementedError()


class StaticChangeNotesProvider(BaseChangeNotesProvider):
    def __init__(self, release_notes_generator, change_notes):
        super(StaticChangeNotesProvider, self).__init__(release_notes_generator)
        self.change_notes = change_notes

    def __call__(self):
        for change_note in self.change_notes:
            yield change_note


class DirectoryChangeNotesProvider(BaseChangeNotesProvider):
    def __init__(self, release_notes_generator, directory):
        super(DirectoryChangeNotesProvider, self).__init__(release_notes_generator)
        self.directory = directory

    def __call__(self):
        for item in sorted(os.listdir(self.directory)):
            yield open("{}/{}".format(self.directory, item)).read()


# For backwards-compatibility
import cumulusci.vcs.github.release_notes.provider as githubProvider
from cumulusci.utils.deprecation import warn_moved


class GithubChangeNotesProvider(githubProvider.GithubChangeNotesProvider):
    """Deprecated: use cumulusci.vcs.github.release_notes.provider.GithubChangeNotesProvider instead"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        warn_moved("cumulusci.vcs.github.release_notes.provider", __name__)



import json
from unittest import mock

import pytest
import responses
from responses.matchers import json_params_matcher

from cumulusci.core.config import ServiceConfig, TaskConfig
from cumulusci.core.exceptions import TaskOptionsError, VcsException
from cumulusci.tasks.github.tests.util_github_api import GithubApiTestMixin
from cumulusci.tasks.vcs import CreateRelease
from cumulusci.tests.util import create_project_config

DUMMY_SHA = "21e04cfe480f5293e2f7103eee8a5cbdb94f7982"


@mock.patch("cumulusci.tasks.vcs.release.time.sleep", mock.Mock())
class TestCreateRelease(GithubApiTestMixin):
    def setup_method(self):
        self.repo_owner = "TestOwner"
        self.repo_name = "TestRepo"
        self.repo_api_url = "https://api.github.com/repos/{}/{}".format(
            self.repo_owner, self.repo_name
        )
        self.project_config = create_project_config(
            self.repo_name,
            self.repo_owner,
            repo_commit="21e04cfe480f5293e2f7103eee8a5cbdb94f7982",
        )
        self.project_config.keychain.set_service(
            "github",
            "test_alias",
            ServiceConfig(
                {
                    "username": "TestUser",
                    "token": "TestPass",
                    "email": "testuser@testdomain.com",
                }
            ),
        )

    @responses.activate
    def test_run_task(self):
        responses.add(
            method=responses.GET,
            url=self.repo_api_url,
            json=self._get_expected_repo(owner=self.repo_owner, name=self.repo_name),
        )
        responses.add(
            method=responses.GET,
            url=self.repo_api_url + "/releases/tags/release/1.0",
            status=404,
        )
        responses.add(
            method=responses.GET,
            url=self.repo_api_url + "/git/ref/tags/release/1.0",
            status=404,
        )
        responses.add(
            method=responses.GET,
            url=self.repo_api_url + f"/commits/{DUMMY_SHA}",
            json=self._get_expected_commit(DUMMY_SHA),
            status=200,
        )
        responses.add(
            method=responses.POST,
            url=self.repo_api_url + "/git/tags",
            json=self._get_expected_tag(
                "release/1.0", "21e04cfe480f5293e2f7103eee8a5cbdb94f7982"
            ),
            status=201,
        )
        responses.add(
            method=responses.POST,
            url=self.repo_api_url + "/git/refs",
            json={},
            status=201,
        )
        responses.add(
            method=responses.POST,
            url=self.repo_api_url + "/releases",
            json=self._get_expected_release("release"),
            status=201,
        )

        task = CreateRelease(
            self.project_config,
            TaskConfig(
                {
                    "options": {
                        "version": "1.0",
                        "version_id": "04t000000000000",
                        "dependencies": [{"namespace": "foo", "version": "1.0"}],
                        "release_content": "foo release",
                        "package_type": "1GP",
                        "tag_prefix": "release/",
                    }
                }
            ),
        )
        task()
        assert {
            "tag_name": "release/1.0",
            "name": "1.0",
            "dependencies": [{"namespace": "foo", "version": "1.0"}],
        } == task.return_values
        # confirm the package_type was recorded in the tag message
        tag_request = json.loads(responses.calls._calls[4].request.body)
        assert "package_type: 1GP" in tag_request["message"]
        # confirm we didn't create a prerelease
        release_request = json.loads(responses.calls._calls[-1].request.body)
        assert not release_request["prerelease"]
        assert release_request["body"] == "foo release"

    @responses.activate
    def test_run_task__release_already_exists(self):
        responses.add(
            method=responses.GET,
            url=self.repo_api_url,
            json=self._get_expected_repo(owner=self.repo_owner, name=self.repo_name),
        )
        responses.add(
            method=responses.GET,
            url=self.repo_api_url + "/releases/tags/release/1.0",
            json=self._get_expected_release("release/1.0"),
        )

        task = CreateRelease(
            self.project_config,
            TaskConfig(
                {
                    "options": {
                        "version": "1.0",
                        "version_id": "04t000000000000",
                        "package_type": "1GP",
                        "tag_prefix": "release/",
                    }
                }
            ),
        )
        with pytest.raises(VcsException):
            task()

    @responses.activate
    def test_run_task__no_commit(self):
        responses.add(
            method=responses.GET,
            url=self.repo_api_url,
            json=self._get_expected_repo(owner=self.repo_owner, name=self.repo_name),
        )
        responses.add(
            method=responses.GET,
            url=self.repo_api_url + "/releases/tags/release/1.0",
            status=404,
        )
        del self.project_config._repo_info["commit"]

        with pytest.raises(VcsException):
            CreateRelease(
                self.project_config,
                TaskConfig({"options": {"version": "1.0", "commit": None}}),
            )

    @responses.activate
    def test_run_task__short_commit(self):
        responses.add(
            method=responses.GET,
            url=self.repo_api_url,
            json=self._get_expected_repo(owner=self.repo_owner, name=self.repo_name),
        )
        responses.add(
            method=responses.GET,
            url=self.repo_api_url + "/releases/tags/release/1.0",
            status=404,
        )
        self.project_config._repo_info["commit"] = "too_short"

        with pytest.raises(TaskOptionsError):
            CreateRelease(
                self.project_config, TaskConfig({"options": {"version": "1.0"}})
            )

    @responses.activate
    def test_run_task__with_custom_prefix(self):
        responses.add(
            method=responses.GET,
            url=self.repo_api_url,
            json=self._get_expected_repo(owner=self.repo_owner, name=self.repo_name),
        )
        responses.add(
            method=responses.GET,
            url=self.repo_api_url + "/releases/tags/custom/1.0",
            status=404,
        )
        responses.add(
            method=responses.GET,
            url=self.repo_api_url + "/git/ref/tags/custom/1.0",
            status=404,
        )
        responses.add(
            method=responses.GET,
            url=self.repo_api_url + f"/commits/{DUMMY_SHA}",
            json=self._get_expected_commit(DUMMY_SHA),
            status=200,
        )
        responses.add(
            method=responses.POST,
            url=self.repo_api_url + "/git/tags",
            json=self._get_expected_tag(
                "release/1.0", "21e04cfe480f5293e2f7103eee8a5cbdb94f7982"
            ),
            status=201,
        )
        responses.add(
            method=responses.POST,
            url=self.repo_api_url + "/git/refs",
            json={},
            status=201,
        )
        responses.add(
            method=responses.POST,
            url=self.repo_api_url + "/releases",
            json=self._get_expected_release("release"),
            status=201,
        )

        task = CreateRelease(
            self.project_config,
            TaskConfig(
                {
                    "options": {
                        "version": "1.0",
                        "version_id": "04t000000000000",
                        "dependencies": [{"namespace": "foo", "version": "1.0"}],
                        "package_type": "2GP",
                        "tag_prefix": "custom/",
                    }
                }
            ),
        )
        task()
        assert {
            "tag_name": "custom/1.0",
            "name": "1.0",
            "dependencies": [{"namespace": "foo", "version": "1.0"}],
        } == task.return_values
        assert "package_type: 2GP" in responses.calls._calls[4].request.body

    @responses.activate
    def test_run_task__beta_1gp(self):
        responses.add(
            method=responses.GET,
            url=self.repo_api_url,
            json=self._get_expected_repo(owner=self.repo_owner, name=self.repo_name),
        )
        responses.add(
            method=responses.GET,
            url=self.repo_api_url + "/releases/tags/beta/1.0-Beta_1",
            status=404,
        )
        responses.add(
            method=responses.GET,
            url=self.repo_api_url + "/git/ref/tags/beta/1.0-Beta_1",
            status=404,
        )
        responses.add(
            method=responses.GET,
            url=self.repo_api_url + f"/commits/{DUMMY_SHA}",
            json=self._get_expected_commit(DUMMY_SHA),
            status=200,
        )
        responses.add(
            method=responses.POST,
            url=self.repo_api_url + "/git/tags",
            json=self._get_expected_tag(
                "beta/1.0-Beta_1", "21e04cfe480f5293e2f7103eee8a5cbdb94f7982"
            ),
            status=201,
        )
        responses.add(
            method=responses.POST,
            url=self.repo_api_url + "/git/refs",
            json={},
            status=201,
        )
        responses.add(
            method=responses.POST,
            url=self.repo_api_url + "/releases",
            json=self._get_expected_release("release"),
            status=201,
        )

        task = CreateRelease(
            self.project_config,
            TaskConfig(
                {
                    "options": {
                        "version": "1.0 (Beta 1)",
                        "version_id": "04t000000000000",
                        "package_type": "1GP",
                        "tag_prefix": "beta/",
                    }
                }
            ),
        )
        task()
        assert {
            "tag_name": "beta/1.0-Beta_1",
            "name": "1.0 (Beta 1)",
            "dependencies": [],
        } == task.return_values
        # confirm we didn't create a prerelease
        release_request = json.loads(responses.calls._calls[-1].request.body)
        assert release_request["prerelease"]

    @responses.activate
    def test_run_task__with_beta_2gp(self):
        responses.add(
            method=responses.GET,
            url=self.repo_api_url,
            json=self._get_expected_repo(owner=self.repo_owner, name=self.repo_name),
        )
        responses.add(
            method=responses.GET,
            url=self.repo_api_url + "/releases/tags/beta/1.1",
            status=404,
        )
        responses.add(
            method=responses.GET,
            url=self.repo_api_url + "/git/ref/tags/beta/1.1",
            status=404,
        )
        responses.add(
            method=responses.GET,
            url=self.repo_api_url + f"/commits/{DUMMY_SHA}",
            json=self._get_expected_commit(DUMMY_SHA),
            status=200,
        )
        responses.add(
            method=responses.POST,
            url=self.repo_api_url + "/git/tags",
            json=self._get_expected_tag(
                "release/1.1", "21e04cfe480f5293e2f7103eee8a5cbdb94f7982"
            ),
            status=201,
        )
        responses.add(
            method=responses.POST,
            url=self.repo_api_url + "/git/refs",
            json={},
            status=201,
        )
        responses.add(
            method=responses.POST,
            url=self.repo_api_url + "/releases",
            json=self._get_expected_release("release"),
            match=[
                json_params_matcher(
                    {
                        "tag_name": "beta/1.1",
                        "name": "1.1",
                        "draft": False,
                        "prerelease": True,
                    }
                )
            ],
            status=201,
        )

        task = CreateRelease(
            self.project_config,
            TaskConfig(
                {
                    "options": {
                        "version": "1.1",
                        "version_id": "04t000000000000",
                        "dependencies": [{"namespace": "foo", "version": "1.0"}],
                        "package_type": "2GP",
                        "tag_prefix": "beta/",
                    }
                }
            ),
        )
        task()
        assert {
            "tag_name": "beta/1.1",
            "name": "1.1",
            "dependencies": [{"namespace": "foo", "version": "1.0"}],
        } == task.return_values
        assert "package_type: 2GP" in responses.calls._calls[4].request.body



import logging
import re
from typing import Dict

import github3.exceptions

from cumulusci.core.config import BaseProjectConfig
from cumulusci.core.exceptions import GithubApiNotFoundError
from cumulusci.core.versions import PackageVersionNumber
from cumulusci.tasks.release_notes.parser import ChangeNotesLinesParser, IssuesParser
from cumulusci.utils.yaml.cumulusci_yml import ReleaseNotesParser


def parser_configs(project_config: BaseProjectConfig) -> Dict[int, ReleaseNotesParser]:
    configs = (
        project_config.project__git__release_notes__parsers__github.values()
        if project_config.project__git__release_notes__parsers__github
        else {}
    )
    configs = (
        project_config.project__git__release_notes__parsers.values()
        if not configs and project_config.project__git__release_notes__parsers
        else configs
    )
    return configs


class GithubLinesParser(ChangeNotesLinesParser):
    def __init__(self, release_notes_generator, title):
        super(GithubLinesParser, self).__init__(release_notes_generator, title)
        self.link_pr = release_notes_generator.link_pr
        self.pr_number = None
        self.pr_url = None

    def _process_change_note(self, pull_request):
        self.pr_number = pull_request.number
        self.pr_url = pull_request.html_url
        return pull_request.body

    def _add_link(self, line):
        if self.link_pr:
            line += " [[PR{}]({})]".format(self.pr_number, self.pr_url)
        return line


class GithubIssuesParser(IssuesParser):
    ISSUE_COMMENT = {
        "beta": "Included in beta release",
        "prod": "Included in production release",
    }

    def __new__(cls, release_notes_generator, title, issue_regex=None):
        if not release_notes_generator.has_issues:
            logging.getLogger(__file__).warning(
                "Issues are disabled for this repository. Falling back to change notes parser."
            )
            return GithubLinesParser(release_notes_generator, title)

        return super().__new__(cls)

    def __init__(self, release_notes_generator, title, issue_regex=None):
        super().__init__(release_notes_generator, title, issue_regex)
        self.link_pr = release_notes_generator.link_pr
        self.pr_number = None
        self.pr_url = None
        self.publish = release_notes_generator.do_publish
        self.github = release_notes_generator.github

    def _add_line(self, line):
        # find issue numbers per line
        issue_numbers = re.findall(self.issue_regex, line, flags=re.IGNORECASE)
        for issue_number in issue_numbers:
            self.content.append(
                {
                    "issue_number": int(issue_number),
                    "pr_number": self.pr_number,
                    "pr_url": self.pr_url,
                }
            )

    def _get_default_regex(self):
        keywords = (
            "close",
            "closes",
            "closed",
            "fix",
            "fixes",
            "fixed",
            "resolve",
            "resolves",
            "resolved",
        )
        return r"(?:{})\s\[?#(\d+)\]?".format("|".join(keywords))

    def _render_content(self):
        content = []
        for item in sorted(self.content, key=lambda k: k["issue_number"]):
            issue = self._get_issue(item["issue_number"])
            txt = "#{}: {}".format(item["issue_number"], issue.title)
            if self.link_pr:
                txt += " [[PR{}]({})]".format(item["pr_number"], item["pr_url"])
            content.append(txt)
            if self.publish:
                self._add_issue_comment(issue)
        return "\r\n".join(content)

    def _get_issue(self, issue_number):
        try:
            issue = self.github.issue(
                self.release_notes_generator.github_info["github_owner"],
                self.release_notes_generator.github_info["github_repo"],
                issue_number,
            )
        except github3.exceptions.NotFoundError:
            raise GithubApiNotFoundError("Issue #{} not found".format(issue_number))
        return issue

    def _process_change_note(self, pull_request):
        self.pr_number = pull_request.number
        self.pr_url = pull_request.html_url
        return pull_request.body

    def _add_issue_comment(self, issue):
        # Ensure all issues have a comment on which release they were fixed
        prefix_beta = self.release_notes_generator.github_info["prefix_beta"]
        prefix_prod = self.release_notes_generator.github_info["prefix_prod"]

        # ParentPullRequestNotesGenerator doesn't utilize a current_tag
        if not hasattr(self.release_notes_generator, "current_tag"):
            return
        elif self.release_notes_generator.current_tag.startswith(prefix_beta):
            is_beta = True
        elif self.release_notes_generator.current_tag.startswith(prefix_prod):
            is_beta = False
        else:
            # not production or beta tag, don't comment
            return
        if is_beta:
            comment_prefix = self.ISSUE_COMMENT["beta"]
        else:
            comment_prefix = self.ISSUE_COMMENT["prod"]
        version_str = PackageVersionNumber.parse_tag(
            self.release_notes_generator.current_tag, prefix_beta, prefix_prod
        ).format()
        has_comment = False
        for comment in issue.comments():
            if comment.body.startswith(comment_prefix):
                has_comment = True
                break
        if not has_comment:
            issue.create_comment("{} {}".format(comment_prefix, version_str))



from cumulusci.core.utils import process_bool_arg
from cumulusci.tasks.base_source_control_task import BaseSourceControlTask
from cumulusci.utils.git import is_release_branch
from cumulusci.vcs.models import AbstractBranch, AbstractRepoCommit


class MergeBranch(BaseSourceControlTask):
    task_docs = """
    Merges the most recent commit on the current branch into other branches depending on the value of source_branch.

    If source_branch is a branch that does not start with the specified branch_prefix, then the commit will be
    merged to all branches that begin with branch_prefix and are not themselves child branches (i.e. branches don't contain '__' in their name).

    If source_branch begins with branch_prefix, then the commit is merged to all child branches of source_branch.
    """
    task_options = {  # TODO: should use `class Options instead`
        "commit": {
            "description": "The commit to merge into feature branches.  Defaults to the current head commit."
        },
        "source_branch": {
            "description": "The source branch to merge from.  Defaults to project__git__default_branch."
        },
        "branch_prefix": {
            "description": "A list of prefixes of branches that should receive the merge.  Defaults to project__git__prefix_feature"
        },
        "skip_future_releases": {
            "description": "If true, then exclude branches that start with the branch prefix if they are not for the lowest release number. Defaults to True."
        },
        "update_future_releases": {
            "description": "If true, then include release branches that are not the lowest release number even if they are not child branches. Defaults to False."
        },
        "create_pull_request_on_conflict": {
            "description": "If true, then create a pull request when a merge conflict arises. Defaults to True."
        },
    }

    def _init_options(self, kwargs):
        super()._init_options(kwargs)

        if "commit" not in self.options:
            self.options["commit"] = self.project_config.repo_commit
        if "branch_prefix" not in self.options:
            self.options[
                "branch_prefix"
            ] = self.project_config.project__git__prefix_feature
        if "source_branch" not in self.options:
            self.options[
                "source_branch"
            ] = self.project_config.project__git__default_branch
        if "skip_future_releases" not in self.options:
            self.options["skip_future_releases"] = True
        else:
            self.options["skip_future_releases"] = process_bool_arg(
                self.options.get("skip_future_releases")
            )
        self.options["update_future_releases"] = process_bool_arg(
            self.options.get("update_future_releases") or False
        )
        if "create_pull_request_on_conflict" not in self.options:
            self.options["create_pull_request_on_conflict"] = True
        else:
            self.options["create_pull_request_on_conflict"] = process_bool_arg(
                self.options.get("create_pull_request_on_conflict")
            )

    def _init_task(self):
        super()._init_task()
        self.repo = self.get_repo()

    def _run_task(self):
        self._validate_source_branch(self.options["source_branch"])
        branches_to_merge = self._get_branches_to_merge()

        for branch in branches_to_merge:
            self._merge(
                branch.name,
                self.options["source_branch"],
                self.options["commit"],
            )

    def _validate_source_branch(self, source_branch):
        """Validates that the source branch exists in the repository"""
        branch: AbstractBranch = self.repo.branch(source_branch)
        self.options["source_branch"] = branch.name

    def _get_existing_prs(self, source_branch, branch_prefix):
        """Returns the existing pull requests from the source branch
        to other branches that are candidates for merging."""
        existing_prs = []
        for pr in self.repo.pull_requests(state="open"):
            if pr.base_ref.startswith(branch_prefix) and pr.head_ref == source_branch:
                existing_prs.append(pr.base_ref)
        return existing_prs

    def _get_branches_to_merge(self):
        """
        If source_branch is the default branch (or a branch that doesn't start with a prefix), we
        gather all branches with branch_prefix that are not child branches.
        NOTE: We only include the _next_ closest release branch when automerging from main.
        A change on main may conflict with the current contents of the lowest release branch.
        In this case, we would like for that conflict to only need to be resolved once
        (not once for each release branch).

        If source_branch starts with branch prefix, we gather
        all branches with branch_prefix that are direct descendents of source_branch.

        If update_future_releases is True, and source_branch is a release branch
        then we also collect all future release branches.
        """
        repo_branches = list(self.repo.branches())
        next_release = self._get_next_release(repo_branches)
        skip_future_releases = self.options["skip_future_releases"]
        update_future_releases = self._update_future_releases(next_release)

        child_branches = []
        main_descendents = []
        release_branches = []
        for branch in repo_branches:
            # check for adding future release branches
            if update_future_releases and self._is_future_release_branch(
                branch.name, next_release
            ):
                release_branches.append(branch)
                continue

            # check if we looking at the source_branch
            if branch.name == self.options["source_branch"]:
                self.logger.debug(f"Skipping branch {branch.name}: is source branch")
                continue

            # check for branch prefix match
            elif not branch.name.startswith(self.options["branch_prefix"]):
                self.logger.debug(
                    f"Skipping branch {branch.name}: does not match prefix '{self.options['branch_prefix']}'"
                )
                continue

            # check if source_branch doesn't have prefix and is not a child (e.g. main)
            elif (
                not self.options["source_branch"].startswith(
                    self.options["branch_prefix"]
                )
                and "__" not in branch.name
            ):
                # only merge to the lowest numbered release branch
                # when merging from a branch without a prefix (e.g. main)
                if skip_future_releases and self._is_future_release_branch(
                    branch.name, next_release
                ):
                    continue
                main_descendents.append(branch)

            # else, we have a branch that starts with branch_prefix
            # check is this branch is a direct descendent
            elif self._is_source_branch_direct_descendent(branch.name):
                child_branches.append(branch)

            # else not a direct descendent
            else:
                self.logger.debug(
                    f"Skipping branch {branch.name}: is not a direct descendent of {self.options['source_branch']}"
                )

        to_merge = []
        if child_branches:
            self.logger.debug(
                f"Found child branches to update: {[branch.name for branch in child_branches]}"
            )
            to_merge = child_branches
        elif self.options["source_branch"].startswith(self.options["branch_prefix"]):
            self.logger.debug(
                f"No children found for branch {self.options['source_branch']}"
            )

        if release_branches:
            self.logger.debug(
                f"Found future release branches to update: {[branch.name for branch in release_branches]}"
            )
            to_merge = to_merge + release_branches

        if main_descendents:
            self.logger.debug(
                f"Found descendents of {self.options['source_branch']} to update: {[branch.name for branch in main_descendents]}"
            )
            to_merge = to_merge + main_descendents

        return to_merge

    def _get_next_release(self, repo_branches):
        """Returns the integer that corresponds to the lowest release number found on all release branches.
        NOTE: We assume that once a release branch is merged that it will be deleted.
        """
        release_nums = [
            self._get_release_number(branch.name)
            for branch in repo_branches
            if self._is_release_branch(branch.name)
        ]
        next_release = sorted(release_nums)[0] if release_nums else None
        return next_release

    def _update_future_releases(self, next_release):
        """Determines whether or not to update future releases.
        Returns True if all of the below checks are True. False otherwise.

        Checks:
        (1) Did we receive the 'update_future_release' flag?
        (2) Is the source_branch a release branch?
        (3) Is it the lowest numbered release branch that exists?

        NOTE: This functionality assumes that the lowest numbered release branch in the repo is
        the next closest release. Put another way, once a release branch is merged we assume that it is immediately deleted.
        """
        update_future_releases = False
        if (
            self.options["update_future_releases"]
            and self._is_release_branch(self.options["source_branch"])
            and next_release == self._get_release_number(self.options["source_branch"])
        ):
            update_future_releases = True
        return update_future_releases

    def _is_release_branch(self, branch_name):
        """A release branch begins with the given prefix"""
        return is_release_branch(branch_name, self.options["branch_prefix"])

    def _get_release_number(self, branch_name) -> int:
        """Get the release number from a release branch name.

        Assumes we already know it is a release branch.
        """
        return int(branch_name.split(self.options["branch_prefix"])[1])

    def _merge(self, branch_name, source, commit):
        """Attempt to merge a commit from source to branch with branch_name"""
        compare = self.repo.compare_commits(branch_name, commit, source)
        if not compare or not compare.files:
            self.logger.info(f"Skipping branch {branch_name}: no file diffs found")
            return

        ret = self.repo.merge(branch_name, commit, source)
        if isinstance(ret, AbstractRepoCommit):
            self.logger.info(
                f"Merged {compare.behind_by} commits into branch: {branch_name}"
            )
        elif ret is None:
            if self.options["create_pull_request_on_conflict"]:
                self._create_conflict_pull_request(branch_name, source)
            else:
                self.logger.info(
                    f"Merge conflict on branch {branch_name}: skipping pull request creation"
                )

    def _create_conflict_pull_request(self, branch_name, source):
        """Attempt to create a pull request from source into branch_name if merge operation encounters a conflict"""
        if branch_name in self._get_existing_prs(
            self.options["source_branch"], self.options["branch_prefix"]
        ):
            self.logger.info(
                f"Merge conflict on branch {branch_name}: merge PR already exists"
            )
            return

        pull = self.repo.create_pull(
            title=f"Merge {source} into {branch_name}",
            base=branch_name,
            head=source,
            body="This pull request was automatically generated because "
            "an automated merge hit a merge conflict",
            options={
                "error_message": f"Error creating merge conflict pull request to merge {source} into {branch_name}"
            },
        )

        if pull is not None:
            self.logger.info(
                f"Merge conflict on branch {branch_name}: created pull request #{pull.number}"
            )
            return pull

    def _is_source_branch_direct_descendent(self, branch_name):
        """Returns True if branch is a direct descendent of the source branch"""
        source_dunder_count = self.options["source_branch"].count("__")
        return (
            branch_name.startswith(f"{self.options['source_branch']}__")
            and branch_name.count("__") == source_dunder_count + 1
        )

    def _is_future_release_branch(self, branch_name, next_release):
        return (
            self._is_release_branch(branch_name)
            and branch_name != self.options["source_branch"]
            and self._get_release_num(branch_name) > next_release
        )

    def _get_release_num(self, release_branch_name):
        """Given a release branch, returns an integer that
        corresponds to the release number for that branch"""
        return int(release_branch_name.split(self.options["branch_prefix"])[1])



from io import StringIO
from unittest.mock import MagicMock, patch

import pytest
from requests.exceptions import ConnectionError

from cumulusci.salesforce_api.retrieve_profile_api import RetrieveProfileApi
from cumulusci.tasks.salesforce.BaseSalesforceApiTask import BaseSalesforceApiTask
from cumulusci.utils.parallel.queries_in_parallel.run_queries_in_parallel import (
    RunParallelQueries,
)


@pytest.fixture
def retrieve_profile_api_instance():
    sf_mock = MagicMock()
    bulk_mock = MagicMock()
    project_config = MagicMock()
    task_config = MagicMock()
    org_config = MagicMock()
    project_config.config = {"project": {"package": {"api_version": "58.0"}}}
    sf_mock.query.return_value = {"records": []}
    api = RetrieveProfileApi(
        project_config=project_config, org_config=org_config, task_config=task_config
    )
    api.sf = sf_mock
    api.bulk = bulk_mock
    return api


def test_init_task(retrieve_profile_api_instance):
    with patch.object(BaseSalesforceApiTask, "_init_task"):
        retrieve_profile_api_instance._init_task()

    assert retrieve_profile_api_instance.api_version == "58.0"


def test_retrieve_existing_profiles(retrieve_profile_api_instance):
    profiles = ["Profile1", "Profile2", "Admin"]
    result = {"records": [{"Name": "Profile1"}]}
    with patch.object(
        RetrieveProfileApi, "_build_query", return_value="some_query"
    ), patch.object(RetrieveProfileApi, "_run_query", return_value=result):
        existing_profiles = retrieve_profile_api_instance._retrieve_existing_profiles(
            profiles
        )

    assert "Profile1" in existing_profiles
    assert "Profile2" not in existing_profiles
    assert "Admin" in existing_profiles
    assert "System Administrator" in existing_profiles


def test_run_query_sf(retrieve_profile_api_instance):
    query = "SELECT Id FROM Account"
    result_data = {"records": [{"Id": "001abc"}]}
    retrieve_profile_api_instance.sf.query.return_value = result_data

    result = retrieve_profile_api_instance._run_query(query)
    assert result == result_data


def test_run_query_bulk(retrieve_profile_api_instance):
    query = "SELECT Id FROM Account"
    result_data = {"records": [{"Id": "001abc"}]}
    retrieve_profile_api_instance.sf.query.side_effect = ConnectionError

    retrieve_profile_api_instance.bulk.create_query_job.return_value = "job_id"
    retrieve_profile_api_instance.bulk.query.return_value = "batch_id"
    retrieve_profile_api_instance.bulk.wait_for_batch.return_value = None
    retrieve_profile_api_instance.bulk.get_all_results_for_query_batch.return_value = [
        "some_value"
    ]

    with patch(
        "salesforce_bulk.util.IteratorBytesIO",
        return_value=StringIO('[{"Id": "001abc"}]'),
    ):
        result = retrieve_profile_api_instance._run_query(query)

    assert result == result_data


def test_extract_table_name_from_query_valid(retrieve_profile_api_instance):
    query = "SELECT Id FROM Account"
    table_name = retrieve_profile_api_instance._extract_table_name_from_query(query)
    assert table_name == "Account"


def test_extract_table_name_from_query_invalid(retrieve_profile_api_instance):
    query = "SELECT Id"
    with pytest.raises(ValueError):
        retrieve_profile_api_instance._extract_table_name_from_query(query)


def test_build_query_basic(retrieve_profile_api_instance):
    columns = ["Id", "Name"]
    table_name = "Account"
    expected_query = "SELECT Id, Name FROM Account"
    result_query = retrieve_profile_api_instance._build_query(columns, table_name)
    assert result_query == expected_query


def test_build_query_with_where(retrieve_profile_api_instance):
    columns = ["Id", "Name"]
    table_name = "Account"
    where = {"Type": "Customer", "Status": ["Active", "Inactive"]}
    expected_query = "SELECT Id, Name FROM Account WHERE Type = 'Customer' AND Status IN ('Active', 'Inactive')"
    result_query = retrieve_profile_api_instance._build_query(
        columns, table_name, where
    )
    assert result_query == expected_query


def test_build_query_empty_columns(retrieve_profile_api_instance):
    columns = []
    table_name = "Account"
    with pytest.raises(ValueError):
        retrieve_profile_api_instance._build_query(columns, table_name)


def test_build_query_empty_table_name(retrieve_profile_api_instance):
    columns = ["Id", "Name"]
    table_name = ""
    with pytest.raises(ValueError):
        retrieve_profile_api_instance._build_query(columns, table_name)


def test_process_sObject_results(retrieve_profile_api_instance):
    result_list = [
        {"SobjectType": "Account"},
        {"SobjectType": "Contact"},
        {"SobjectType": "Opportunity"},
    ]
    expected_result = {"CustomObject": ["Account", "Contact", "Opportunity"]}
    result = retrieve_profile_api_instance._process_sObject_results(result_list)
    assert result == expected_result


def test_process_sObject_results_missing_key(retrieve_profile_api_instance):
    result_list = [{"ObjectType": "Account"}]
    with pytest.raises(KeyError):
        retrieve_profile_api_instance._process_sObject_results(result_list)


def test_process_customTab_results(retrieve_profile_api_instance):
    result_list = [
        {"Name": "CustomTab1"},
        {"Name": "CustomTab2"},
        {"Name": "CustomTab3"},
    ]
    expected_result = {"CustomTab": ["CustomTab1", "CustomTab2", "CustomTab3"]}
    result = retrieve_profile_api_instance._process_customTab_results(result_list)
    assert result == expected_result


def test_process_customTab_results_missing_key(retrieve_profile_api_instance):
    result_list = [{"TabName": "CustomTab1"}]
    with pytest.raises(KeyError):
        retrieve_profile_api_instance._process_customTab_results(result_list)


def test_process_setupEntityAccess_results(retrieve_profile_api_instance):
    result_list = [
        {"SetupEntityType": "ApexClass", "SetupEntityId": "001abc"},
        {"SetupEntityType": "ApexPage", "SetupEntityId": "002def"},
        {"SetupEntityType": "CustomPermission", "SetupEntityId": "003ghi"},
    ]

    queries_result = {
        "ApexClass": [
            {"Id": "001abc", "Name": "TestApexClass", "NamespacePrefix": "apex"}
        ],
        "ApexPage": [{"Id": "002def", "Name": "TestApexPage"}],
        "CustomPermission": [],
    }
    with patch.object(
        RetrieveProfileApi, "_build_query", return_value="SELECT Id, Name FROM Table"
    ) as mock_build_query, patch.object(
        RunParallelQueries,
        "_run_queries_in_parallel",
        return_value=queries_result,
    ) as mock_run_queries:

        (
            entities,
            result,
        ) = retrieve_profile_api_instance._process_setupEntityAccess_results(
            result_list
        )
        mock_build_query.assert_any_call(
            ["Name", "NamespacePrefix"], "ApexClass", {"Id": ["001abc"]}
        )
        mock_build_query.assert_any_call(
            ["Name", "NamespacePrefix"], "ApexPage", {"Id": ["002def"]}
        )
        mock_build_query.assert_any_call(
            ["DeveloperName", "NamespacePrefix"], "CustomPermission", {"Id": ["003ghi"]}
        )
        mock_run_queries.assert_called_with(
            {
                "ApexClass": "SELECT Id, Name FROM Table",
                "ApexPage": "SELECT Id, Name FROM Table",
                "CustomPermission": "SELECT Id, Name FROM Table",
            },
            retrieve_profile_api_instance._run_query,
        )

    expected_entities = {
        "ApexClass": ["apex__TestApexClass"],
        "ApexPage": ["TestApexPage"],
        "CustomPermission": [],
    }
    assert entities == expected_entities
    assert result == queries_result


def test_process_all_results(retrieve_profile_api_instance):
    result_dict = {
        "setupEntityAccess": "some_result",
        "sObject": "some_result",
        "customTab": "some_result",
        "profileFlow": "some_result",
    }
    with patch.object(
        RetrieveProfileApi,
        "_process_setupEntityAccess_results",
        return_value=(
            {
                "ApexClass": ["TestApexClass"],
                "ApexPage": ["TestApexPage"],
                "FlowDefinition": ["TestFlow"],
            },
            {"FlowDefinition": ["some_result"]},
        ),
    ), patch.object(
        RetrieveProfileApi,
        "_process_sObject_results",
        return_value={"CustomObject": ["TestObject"]},
    ), patch.object(
        RetrieveProfileApi,
        "_process_customTab_results",
        return_value={"CustomTab": ["TestTab"]},
    ), patch.object(
        RetrieveProfileApi,
        "_match_profiles_and_flows",
        return_value={"Profile1": ["Flow1"]},
    ):
        entities, profile_flow = retrieve_profile_api_instance._process_all_results(
            result_dict
        )

    print(entities)
    assert entities["ApexClass"] == ["TestApexClass"]
    assert entities["ApexPage"] == ["TestApexPage"]
    assert entities["FlowDefinition"] == ["TestFlow"]
    assert entities["CustomObject"] == ["TestObject"]
    assert entities["CustomTab"] == ["TestTab"]
    assert profile_flow == {"Profile1": ["Flow1"]}


def test_queries_retrieve_permissions(retrieve_profile_api_instance):
    profiles = ["Profile1", "Profile2"]

    with patch.object(RetrieveProfileApi, "_build_query") as mock_build_query:
        retrieve_profile_api_instance._queries_retrieve_permissions(profiles)

    mock_build_query.assert_any_call(
        ["SetupEntityId", "SetupEntityType"],
        "SetupEntityAccess",
        {"Parent.Profile.Name": profiles},
    )
    mock_build_query.assert_any_call(
        ["SObjectType"], "ObjectPermissions", {"Parent.Profile.Name": profiles}
    )
    mock_build_query.assert_any_call(
        ["Name"], "PermissionSetTabSetting", {"Parent.Profile.Name": profiles}
    )


def test_retrieve_permissionable_entities(retrieve_profile_api_instance):
    profiles = ["Profile1", "Profile2"]
    expected_queries = {"query_name": "query"}
    expected_result = (
        {
            "ApexClass": ["TestApexClass"],
            "ApexPage": ["TestApexPage"],
            "CustomObject": ["TestObject"],
            "CustomTab": ["TestTab"],
        },
        {"Profile1": ["Flow1"]},
    )

    with patch.object(
        RunParallelQueries, "_run_queries_in_parallel"
    ) as mock_run_queries, patch.object(
        RetrieveProfileApi,
        "_queries_retrieve_permissions",
        return_value=expected_queries,
    ), patch.object(
        RetrieveProfileApi, "_process_all_results", return_value=expected_result
    ):

        result = retrieve_profile_api_instance._retrieve_permissionable_entities(
            profiles
        )
        mock_run_queries.assert_called_with(
            expected_queries, retrieve_profile_api_instance._run_query
        )

        assert result == expected_result


def test_match_profiles_and_flows(retrieve_profile_api_instance):
    result_profiles = [
        {"SetupEntityId": "001abc", "Parent": {"Profile": {"Name": "Profile1"}}},
        {"SetupEntityId": "001abc", "Parent": {"Profile": {"Name": "Profile2"}}},
        {"SetupEntityId": "002def", "Parent": {"Profile": {"Name": "Profile1"}}},
    ]

    result_flows = [
        {"ApiName": "Flow1", "attributes": {"url": "instance_url/001abc"}},
        {"ApiName": "Flow2", "attributes": {"url": "instance_url/002def"}},
    ]

    profile_flow = retrieve_profile_api_instance._match_profiles_and_flows(
        result_profiles, result_flows
    )

    assert "Flow1" in profile_flow["Profile1"]
    assert "Flow2" in profile_flow["Profile1"]
    assert "Flow1" in profile_flow["Profile2"]
    assert "Flow2" not in profile_flow["Profile2"]


if __name__ == "__main__":
    pytest.main()



import typing as T
from functools import wraps
from unittest import mock
from unittest.mock import call

import pytest

from cumulusci.core import exceptions as exc
from cumulusci.tasks.bulkdata.step import (
    DataApi,
    DataOperationJobResult,
    DataOperationResult,
    DataOperationStatus,
    DataOperationType,
)
from cumulusci.tasks.bulkdata.tests.integration_test_utils import ensure_accounts
from cumulusci.tasks.bulkdata.update_data import UpdateData


def _hashify_operation(kwargs):
    return frozenset(
        {
            **kwargs,
            "context": None,
            "fields": frozenset(kwargs["fields"]),
            "api_options": frozenset(kwargs.get("api_options", {}).items()),
        }.items()
    )


def noop(*args, **kwargs):
    pass


ensure_accounts = ensure_accounts  # cleans up multiple lint errors at once.


class MockBulkAPIResponses:
    def activate(self, func):
        @wraps(func)
        def wrapper(*args, **kwds):
            self.mock_bulk_API_responses_context = MockBulkAPIResponsesContext()
            with mock.patch(
                "cumulusci.tasks.bulkdata.update_data.get_query_operation",
                self.mock_bulk_API_responses_context.get_query_operation,
            ), mock.patch(
                "cumulusci.tasks.bulkdata.update_data.get_dml_operation",
                self.mock_bulk_API_responses_context.get_dml_operation,
            ):
                try:
                    ret = func(*args, **kwds)
                finally:
                    self.mock_bulk_API_responses_context = None
            return ret

        return wrapper

    @property
    def add_query_operation(self):
        return self.mock_bulk_API_responses_context.add_query_operation

    @property
    def add_dml_operation(self):
        return self.mock_bulk_API_responses_context.add_dml_operation


class FakeJobOperationResults(T.NamedTuple):
    results: list
    job_result_status: DataOperationStatus
    loader_callback: T.Callable = noop


class MockBulkAPIResponsesContext:
    def __init__(self):
        self.query_operations = {}
        self.dml_operations = {}

    def add_query_operation(
        self, *, results, job_result_status=DataOperationStatus.SUCCESS, **expected
    ):
        self.query_operations[_hashify_operation(expected)] = FakeJobOperationResults(
            results,
            job_result_status,
        )

    def add_dml_operation(
        self,
        *,
        results,
        loader_callback=noop,
        job_result_status=DataOperationStatus.SUCCESS,
        **expected,
    ):
        val = FakeJobOperationResults(results, job_result_status, loader_callback)
        self.dml_operations[_hashify_operation(expected)] = val

    def get_query_operation(self, **given):
        try:
            results = self.query_operations[_hashify_operation(given)]
        except KeyError:
            raise KeyError(f"Cannot find response matching {given}")
        return FakeOperationResult(results)

    def get_dml_operation(self, **given):
        try:
            results = self.dml_operations[_hashify_operation(given)]
        except KeyError:
            raise KeyError(f"Cannot find response matching {given}")
        return FakeDMLOperationResult(results, fields=given["fields"])


class FakeOperationResult:
    def __init__(self, results: FakeJobOperationResults):
        self.results = results

    def yield_per(self, number):  # TODO
        return self.results.results

    def count(self):
        return len(self.results.results)

    def query(self):
        pass

    @property
    def job_result(self):
        return DataOperationJobResult(
            self.results.job_result_status, [], len(self.results.results), 0
        )

    def get_results(self):
        return self.results.results


class FakeDMLOperationResult(FakeOperationResult):
    def __init__(self, results: FakeJobOperationResults, fields=None):
        self.results = results
        self.fields = fields

    def start(self):
        pass

    def load_records(self, records, *args, **kwargs):
        for record in records:
            self.results.loader_callback(dict(zip(self.fields, record)))

    def end(self):
        pass

    @property
    def job_result(self):
        return DataOperationJobResult(
            self.results.job_result_status,
            [],
            len(self.results.results),
            sum(1 for r in self.results.results if r.success is False),
        )


def _fake_val_and_inject_ns(
    should_inject_namespaces: bool,
    sobjects_to_validate: list,
    operation_to_validate: str,
):
    return sobjects_to_validate


bulkapi_responses = MockBulkAPIResponses()


class TestUpdates:
    def test_options__bad_sobject(self, create_task):
        with pytest.raises(exc.TaskOptionsError) as e:
            task = create_task(UpdateData, {"object": "a b c"})
            task()
        assert "a b c" in str(e.value), e.value

    def test_options__bad_field(self, create_task):
        with pytest.raises(exc.TaskOptionsError) as e:
            task = create_task(UpdateData, {"object": "Account", "fields": ["a b c"]})
            task()
        assert "a b c" in str(e.value), e.value

    def test_options__bad_fields(self, create_task):
        with pytest.raises(exc.TaskOptionsError) as e:
            task = create_task(
                UpdateData, {"object": "Account", "fields": ["a b c", "1 2 3"]}
            )
            task()
        assert "a b c" in str(e.value), e.value

    def test_options__bad_api(self, create_task):
        with pytest.raises(exc.TaskOptionsError) as e:
            task = create_task(UpdateData, {"object": "Account", "api": "XML-RPC"})
            task()
        assert "XML-RPC" in str(e.value), e.value

    @bulkapi_responses.activate
    def test_simple_task(self, create_task):

        bulkapi_responses.add_query_operation(
            sobject="Account",
            fields=["Id", "Name"],
            query="SELECT Id,Name FROM Account",
            api=DataApi.SMART,
            results=[
                ["OID000BLAH", "Mark Benihoff"],
                ["OID000BLAH2", "Parkour Hairish"],
            ],
        )
        loader = mock.Mock()
        bulkapi_responses.add_dml_operation(
            sobject="Account",
            operation=DataOperationType.UPDATE,
            fields=["BillingStreet", "Description", "NumberOfEmployees", "Id"],
            api_options={},
            api=DataApi.SMART,
            volume=2,
            results=[
                DataOperationResult("OID000BLAH", success=True, error=""),
                DataOperationResult("OID000BLAH2", success=True, error=""),
            ],
            loader_callback=loader,
        )

        task = create_task(
            UpdateData,
            {
                "object": "Account",
                "recipe": "datasets/update.recipe.yml",
                "fields": ["Name"],
            },
        )

        task._validate_and_inject_namespace_prefixes = _fake_val_and_inject_ns
        task()
        assert loader.mock_calls == [
            call(
                {
                    "BillingStreet": mock.ANY,
                    "Description": "Mark Benihoff is our favorite customer",
                    "NumberOfEmployees": "10000",
                    "Id": "OID000BLAH",
                }
            ),
            call(
                {
                    "BillingStreet": mock.ANY,
                    "Description": "Parkour Hairish is our favorite customer",
                    "NumberOfEmployees": "10000",
                    "Id": "OID000BLAH2",
                }
            ),
        ], loader.mock_calls

    @bulkapi_responses.activate
    def test_with_no_query_results(self, create_task):
        bulkapi_responses.add_query_operation(
            sobject="Account",
            fields=["Id", "Name"],
            query="SELECT Id,Name FROM Account",
            api=DataApi.SMART,
            results=[],
        )

        task = create_task(
            UpdateData,
            {
                "object": "Account",
                "recipe": "datasets/update.recipe.yml",
                "fields": ["Name"],
            },
        )

        task._validate_and_inject_namespace_prefixes = _fake_val_and_inject_ns
        task.logger = mock.Mock()
        task()
        last_message = task.logger.mock_calls[-1].args[0]
        assert last_message.startswith("No records found"), last_message

    @bulkapi_responses.activate
    def test_with_where_clause(self, create_task):
        bulkapi_responses.add_query_operation(
            sobject="Account",
            fields=["Id", "Name"],
            query="SELECT Id,Name FROM Account WHERE name LIKE '%ark%'",
            api=DataApi.SMART,
            results=[
                ["OID000BLAH", "Mark Benihoff"],
                ["OID000BLAH2", "Parkour Hairish"],
            ],
        )
        bulkapi_responses.add_dml_operation(
            sobject="Account",
            operation=DataOperationType.UPDATE,
            fields=["BillingStreet", "Description", "NumberOfEmployees", "Id"],
            api_options={},
            api=DataApi.SMART,
            volume=2,
            results=[
                DataOperationResult("OID000BLAH", success=True, error=""),
                DataOperationResult("OID000BLAH2", success=True, error=""),
            ],
        )
        task = create_task(
            UpdateData,
            {
                "object": "Account",
                "recipe": "datasets/update.recipe.yml",
                "fields": ["Name"],
                "where": "name LIKE '%ark%'",
            },
        )
        task._validate_and_inject_namespace_prefixes = _fake_val_and_inject_ns
        task.logger = mock.Mock()
        task()
        last_message = task.logger.mock_calls[-1].args[0]
        assert last_message.startswith(
            "Account objects matching \"name like '%ark%'\""
        ), last_message

    @bulkapi_responses.activate
    def test_query_failed(self, create_task):
        bulkapi_responses.add_query_operation(
            sobject="Account",
            fields=["Id", "Name"],
            query="SELECT Id,Name FROM Account",
            api=DataApi.SMART,
            results=[],
            job_result_status=DataOperationStatus.JOB_FAILURE,
        )
        loader = mock.Mock()
        bulkapi_responses.add_dml_operation(
            sobject="Account",
            operation=DataOperationType.UPDATE,
            fields=["BillingStreet", "Description", "NumberOfEmployees", "Id"],
            api_options={},
            api=DataApi.SMART,
            volume=2,
            results=[
                DataOperationResult("OID000BLAH", success=True, error=""),
                DataOperationResult("OID000BLAH2", success=True, error=""),
            ],
            loader_callback=loader,
        )

        task = create_task(
            UpdateData,
            {
                "object": "Account",
                "recipe": "datasets/update.recipe.yml",
                "fields": ["Name"],
            },
        )

        task._validate_and_inject_namespace_prefixes = _fake_val_and_inject_ns
        with pytest.raises(exc.BulkDataException) as e:
            task()

        assert "Unable to query records" in str(e.value)
        assert "Account" in str(e.value)

    @bulkapi_responses.activate
    def test_update_query_failed(self, create_task):
        bulkapi_responses.add_query_operation(
            sobject="Account",
            fields=["Id", "Name"],
            query="SELECT Id,Name FROM Account",
            api=DataApi.SMART,
            results=[
                ["OID000BLAH", "Mark Benihoff"],
                ["OID000BLAH2", "Parkour Hairish"],
            ],
        )
        loader = mock.Mock()
        bulkapi_responses.add_dml_operation(
            sobject="Account",
            operation=DataOperationType.UPDATE,
            fields=["BillingStreet", "Description", "NumberOfEmployees", "Id"],
            api_options={},
            api=DataApi.SMART,
            volume=2,
            results=[],
            loader_callback=loader,
            job_result_status=DataOperationStatus.JOB_FAILURE,
        )

        task = create_task(
            UpdateData,
            {
                "object": "Account",
                "recipe": "datasets/update.recipe.yml",
                "fields": ["Name"],
            },
        )

        task._validate_and_inject_namespace_prefixes = _fake_val_and_inject_ns
        with pytest.raises(exc.BulkDataException) as e:
            task()
        assert "Unable to update records" in str(e.value)

    @bulkapi_responses.activate
    def test_update_row_errors(self, create_task):
        bulkapi_responses.add_query_operation(
            sobject="Account",
            fields=["Id", "Name"],
            query="SELECT Id,Name FROM Account",
            api=DataApi.SMART,
            results=[
                ["OID000BLAH", "Mark Benihoff"],
                ["OID000BLAH2", "Parkour Hairish"],
            ],
        )
        loader = mock.Mock()
        bulkapi_responses.add_dml_operation(
            sobject="Account",
            operation=DataOperationType.UPDATE,
            fields=["BillingStreet", "Description", "NumberOfEmployees", "Id"],
            api_options={},
            api=DataApi.SMART,
            volume=2,
            results=[
                DataOperationResult("OID000BEEF", success=True, error=""),
                DataOperationResult("OID000BLAH", success=False, error="humbug"),
            ],
            loader_callback=loader,
            job_result_status=DataOperationStatus.ROW_FAILURE,
        )

        task = create_task(
            UpdateData,
            {
                "object": "Account",
                "recipe": "datasets/update.recipe.yml",
                "fields": ["Name"],
                "ignore_row_errors": True,
            },
        )

        task._validate_and_inject_namespace_prefixes = _fake_val_and_inject_ns
        task.logger = mock.Mock()
        task()
        assert task.return_values["status"] == DataOperationStatus.ROW_FAILURE
        assert task.return_values["total_row_errors"] == 1
        assert (
            task.logger.mock_calls[-1].args[0]
            == "All account objects processed (2). 1 errors"
        ), task.logger.mock_calls[-1].args[0]

    @bulkapi_responses.activate
    def test_update_row_errors_exception_catching(self, create_task):
        bulkapi_responses.add_query_operation(
            sobject="Account",
            fields=["Id", "Name"],
            query="SELECT Id,Name FROM Account",
            api=DataApi.SMART,
            results=[
                ["OID000BLAH", "Mark Benihoff"],
                ["OID000BLAH2", "Parkour Hairish"],
            ],
        )
        loader = mock.Mock()
        bulkapi_responses.add_dml_operation(
            sobject="Account",
            operation=DataOperationType.UPDATE,
            fields=["BillingStreet", "Description", "NumberOfEmployees", "Id"],
            api_options={},
            api=DataApi.SMART,
            volume=2,
            results=[
                DataOperationResult("OID000BEEF", success=True, error=""),
                DataOperationResult("OID000BLAH", success=False, error="humbug"),
            ],
            loader_callback=loader,
            job_result_status=DataOperationStatus.ROW_FAILURE,
        )

        task = create_task(
            UpdateData,
            {
                "object": "Account",
                "recipe": "datasets/update.recipe.yml",
                "fields": ["Name"],
                "ignore_row_errors": False,
            },
        )

        task._validate_and_inject_namespace_prefixes = _fake_val_and_inject_ns
        with pytest.raises(exc.BulkDataException) as e:
            task()
        assert str(e.value) == "1 update error"


class TestUpdatesIntegrationTests:

    # VCR doesn't match because of randomized data
    @pytest.mark.vcr()
    def test_updates_task(self, create_task, ensure_accounts):
        with ensure_accounts(6):
            task = create_task(
                UpdateData,
                {
                    "object": "Account",
                    "recipe": "datasets/update.recipe.yml",
                    "fields": ["Name"],
                },
            )
            task.logger = mock.Mock()
            task()
            last_message = task.logger.mock_calls[-1].args[0]
            assert (
                str(last_message) == "All account objects successfully updated (6)."
            ), str(last_message)



from pathlib import Path
from shutil import copytree, rmtree

from cumulusci.core.exceptions import TaskOptionsError
from cumulusci.core.tasks import BaseTask
from cumulusci.utils import find_replace, update_tree


class CreateManagedSrc(BaseTask):
    task_docs = """
        Apex classes which use the @deprecated annotation can comment
        it out using //cumulusci-managed so that it can be deployed as
        part of unmanaged metadata, where this annotation is not allowed.
        This task is for use when deploying to a packaging org to
        remove the comment so that the annotation takes effect.
    """

    task_options = {
        "path": {
            "description": "The path containing metadata to process for managed deployment",
            "required": True,
        },
        "revert_path": {
            "description": "The path to copy the original metadata to for the revert call",
            "required": True,
        },
    }

    managed_token = "//cumulusci-managed"

    def _run_task(self):
        # Check that path exists
        path = Path(self.options["path"])
        if not path.is_dir():
            raise TaskOptionsError(
                f"The path {path} does not exist or is not a directory"
            )

        # Check that revert_path does not exist
        revert_path = Path(self.options["revert_path"])
        if revert_path.exists():
            raise TaskOptionsError(
                f"The revert_path {self.options['revert_path']} already exists.  Delete it and try again"
            )

        # Copy path to revert_path
        copytree(str(path), str(revert_path))

        # Edit metadata in path
        self.logger.info(
            f"Removing the string {self.managed_token} from {path}/classes and {path}/triggers"
        )
        find_replace(
            self.managed_token,
            "",
            str(path / "classes"),
            "*.cls",
            self.logger,
        )
        find_replace(
            self.managed_token,
            "",
            str(path / "triggers"),
            "*.trigger",
            self.logger,
        )

        self.logger.info(
            f"{self.managed_token} has been stripped from all classes and triggers in {path}"
        )


class RevertManagedSrc(BaseTask):
    task_options = {
        "path": {
            "description": "The path containing metadata to process for managed deployment",
            "required": True,
        },
        "revert_path": {
            "description": "The path to copy the original metadata to for the revert call",
            "required": True,
        },
    }

    def _run_task(self):
        path = Path(self.options["path"])
        revert_path = Path(self.options["revert_path"])
        # Check that revert_path does exists
        if not revert_path.is_dir():
            raise TaskOptionsError(
                f"The revert_path {revert_path} does not exist or is not a directory"
            )

        self.logger.info(f"Reverting {path} from {revert_path}")
        update_tree(str(revert_path), str(path))
        self.logger.info(f"{path} is now reverted")

        # Delete the revert_path
        self.logger.info(f"Deleting {str(revert_path)}")
        rmtree(revert_path)



"""Calls to the MetaDeploy REST API."""

import functools
from typing import Optional, Union

from requests.models import Response
from requests.sessions import Session

from cumulusci.core.config import ServiceConfig
from cumulusci.core.exceptions import CumulusCIException
from cumulusci.utils.http.requests_utils import safe_json_from_response


def make_api_session(metadeploy_service: ServiceConfig) -> Session:
    """
    Given a MetaDeploy ServiceConfig, returns a requests.Session with
    authorization headers and service.base_url set.
    """
    base_url: str = metadeploy_service.url

    patched_session = Session()
    patched_session.headers["Authorization"] = f"token {metadeploy_service.token}"

    def patched_request(base, func, method, url, *args, **kwargs):
        """Call requests.request with base_url prepended."""
        base_url = base + url.replace(base, "", 1)
        return func(method, base_url, *args, **kwargs)

    patched_session.request = functools.partial(
        patched_request, base_url, patched_session.request
    )

    return patched_session


class MetaDeployAPI:
    def __init__(self, metadeploy_service: ServiceConfig) -> None:
        self.session: Session = make_api_session(metadeploy_service=metadeploy_service)

    def _create(self, obj: str, json: dict) -> dict:
        response: Response = self.session.post(obj, json=json)
        return safe_json_from_response(response)

    def _find(self, obj: str, query: Union[dict, str]) -> Optional[dict]:
        response: Response = self.session.get(obj, params=query)
        result: dict = safe_json_from_response(response)

        try:
            return result["data"][0]
        except IndexError:
            return None
        except KeyError:
            raise CumulusCIException(
                "CumulusCI received an unexpected response from MetaDeploy. "
                "Ensure that your MetaDeploy service is configured with the Admin API URL, which "
                "ends in /rest, and that your authentication token is valid."
            ) from None

    def create_plan(self, plan: dict) -> dict:
        return self._create("/plans", plan)

    def create_plan_template(self, template: dict) -> dict:
        return self._create("/plantemplates", json=template)

    def create_plan_slug(self, slug: dict) -> dict:
        return self._create("/planslug", json=slug)

    def create_version(self, version: dict) -> dict:
        return self._create("/versions", json=version)

    def find_product(self, repo_url: str) -> Optional[dict]:
        return self._find("/products", {"repo_url": repo_url})

    def find_plan_template(self, query: dict) -> Optional[dict]:
        return self._find("/plantemplates", query)

    def find_version(self, query: dict) -> Optional[dict]:
        return self._find("/versions", query)

    def update_version(self, pk: Union[int, str]) -> dict:
        response: Response = self.session.patch(
            f"/versions/{pk}", json={"is_listed": True}
        )
        return safe_json_from_response(response)

    def update_lang_translation(self, lang: str, labels: dict) -> dict:
        response: Response = self.session.patch(f"/translations/{lang}", json=labels)
        return safe_json_from_response(response)



import gzip
import re
import typing as T
from collections import defaultdict
from contextlib import ExitStack, contextmanager
from enum import Enum
from logging import getLogger
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Dict, Iterable, List, NamedTuple, Optional

from sqlalchemy import MetaData, create_engine, not_
from sqlalchemy.engine import Engine
from sqlalchemy.orm import create_session, exc, sessionmaker

from cumulusci.salesforce_api.filterable_objects import NOT_COUNTABLE, NOT_EXTRACTABLE
from cumulusci.salesforce_api.org_schema_models import (
    Base,
    Field,
    FileMetadata,
    SObject,
)
from cumulusci.utils.fileutils import FSResource
from cumulusci.utils.http.multi_request import (
    RECOVERABLE_ERRORS,
    CompositeParallelSalesforce,
)
from cumulusci.utils.salesforce.count_sobjects import count_sobjects

y2k = "Sat, 1 Jan 2000 00:00:01 GMT"


def zip_database(tempfile: Path, schema_path: T.Union[FSResource, Path]):
    """Compress tempfile.db to schema_path.db.gz"""
    with tempfile.open("rb") as db:
        with schema_path.open("wb") as fileobj:
            with gzip.GzipFile(fileobj=fileobj, mode="w") as gzipped:
                gzipped.write(db.read())


def unzip_database(gzipfile, outfile):
    """Decompress schema_path.db.gz to outfile.db"""
    with gzipfile.open("rb") as fileobj:
        with gzip.GzipFile(fileobj=fileobj) as gzipped:
            with open(outfile, "wb") as db:
                db.write(gzipped.read())


def ignore_based_on_name(objname, patterns: T.Sequence[re.Pattern]):
    return any(pat.fullmatch(objname) for pat in patterns)


# TODO: Rename this SObjectFilters
class Filters(Enum):
    """Options for filtering schemas"""

    # These were originally done using SQL Alchemy syntax because filtering
    # was done in SQL Alchemy. It is still possible to use them that way
    # in client code:
    #
    # schema.objects.filter(Filters.activateable)
    activateable = SObject.activateable
    compactLayoutable = SObject.compactLayoutable
    createable = SObject.createable
    deepCloneable = SObject.deepCloneable
    deletable = SObject.deletable
    layoutable = SObject.layoutable
    listviewable = SObject.listviewable
    lookupLayoutable = SObject.lookupLayoutable
    mergeable = SObject.mergeable
    queryable = SObject.queryable
    replicateable = SObject.replicateable
    retrieveable = SObject.retrieveable
    searchLayoutable = SObject.searchLayoutable
    searchable = SObject.searchable
    triggerable = SObject.triggerable
    undeletable = SObject.undeletable
    updateable = SObject.updateable

    not_activateable = not_(SObject.activateable)
    not_compactLayoutable = not_(SObject.compactLayoutable)
    not_createable = not_(SObject.createable)
    not_deepCloneable = not_(SObject.deepCloneable)
    not_deletable = not_(SObject.deletable)
    not_layoutable = not_(SObject.layoutable)
    not_listviewable = not_(SObject.listviewable)
    not_lookupLayoutable = not_(SObject.lookupLayoutable)
    not_mergeable = not_(SObject.mergeable)
    not_queryable = not_(SObject.queryable)
    not_replicateable = not_(SObject.replicateable)
    not_retrieveable = not_(SObject.retrieveable)
    not_searchLayoutable = not_(SObject.searchLayoutable)
    not_searchable = not_(SObject.searchable)
    not_triggerable = not_(SObject.triggerable)
    not_undeletable = not_(SObject.undeletable)
    not_updateable = not_(SObject.updateable)

    extractable = "extractable"  # can it be extracted safely?
    # non-extractable objects are discovered
    # through experimentation.
    # Also, all non-queryable and non-retrievable objects are
    # considered non-extractable.

    populated = SObject.count > 0  # does it have data in the org?


# TODO: Profiling and optimizing of the
#      SQL parts. After the object is frozen,
#      all query-sets can be cached as
#      dicts and lists
class Schema:
    """Represents an org's schema, cached from describe() calls"""

    included_objects = None
    includes_counts = False

    def __init__(self, engine, schema_path, filters: T.Sequence[Filters] = ()):
        self.engine = engine
        Session = sessionmaker(bind=self.engine)
        self.session = Session()
        self.path = schema_path
        self.filters = set(filters)

    @property
    def sobjects(self):
        query = self.session.query(SObject)
        if self.included_objects is not None:
            query = query.filter(SObject.name.in_(self.included_objects))
        return query

    def __getitem__(self, name):
        try:
            return self.sobjects.filter_by(name=name).one()
        except exc.NoResultFound:
            raise KeyError(f"No sobject named `{name}`")

    def __contains__(self, name):
        return bool(self.sobjects.filter_by(name=name).first())

    def keys(self):
        return [x.name for x in self.sobjects.all()]

    def values(self):
        return self.sobjects.all()

    def items(self):
        return [(obj.name, obj) for obj in self.sobjects]

    def get(self, name: str):
        return self.sobjects.filter_by(name=name).first()

    def block_writing(self):
        """After this method is called, the database can't be updated again"""
        # changes don't get saved back to the gzip
        # so there is no point writing to the DB
        def closed():
            raise IOError("Database is not open for writing")

        self.session._real_commit__ = self.session.commit
        self.session.commit = closed

    def close(self):
        self.session.close()

    def __repr__(self):
        return f"<Schema {self.path} : {self.engine}>"

    def add_counts(self, counts: T.Dict[str, int]):
        for objname, count in counts.items():
            obj = self.get(objname)
            if obj:
                obj.count = count
        self.includes_counts = True

    def populate_cache(
        self,
        sf,
        included_objects,
        filters: T.Sequence[Filters] = (),
        patterns_to_ignore: T.Sequence[str] = (),
        logger=None,
        *,
        include_counts: bool = False,
    ) -> T.Union[T.Dict[str, int], T.Dict[str, None]]:
        """Populate a schema cache from the API, using last_modified_date
        to pull down only new schema"""
        for pat in patterns_to_ignore:
            assert pat.replace("%", "").isidentifier(), f"Pattern has wrong chars {pat}"

        # Platform bug!
        patterns_to_ignore = ("MacroInstruction%",) + tuple(patterns_to_ignore)
        regexps_to_ignore = [
            re.compile(pat.replace("%", ".*"), re.IGNORECASE)
            for pat in patterns_to_ignore
        ]

        objs = [obj for obj in sf.describe()["sobjects"]]
        if included_objects:
            objs = [obj for obj in objs if obj["name"] in included_objects]
        sobj_names = [
            obj["name"]
            for obj in objs
            if not (
                ignore_based_on_name(obj["name"], regexps_to_ignore)
                or ignore_based_on_properties(obj, filters)
            )
        ]

        def find_last_modified_date(sobj_name: str) -> Optional[str]:
            if obj := self.get(sobj_name):
                return obj.last_modified_date
            else:
                return y2k

        objs_to_refresh = {
            sobj_name: find_last_modified_date(sobj_name) for sobj_name in sobj_names
        }
        changes = list(deep_describe(sf, objs_to_refresh, logger))

        self._populate_cache_from_describe(changes)
        if include_counts:
            results = populate_counts(sf, self, sobj_names, logger)
        else:
            results = {name: None for name in sobj_names}
        return results

    def _populate_cache_from_describe(self, describe_objs: List["DescribeUpdate"]):
        """Populate a schema cache from a list of describe objects."""
        engine = self.engine
        metadata = Base.metadata
        metadata.bind = engine
        metadata.reflect()

        with BufferedSession(engine, metadata) as sess:

            for (sobj_data, last_modified) in describe_objs:
                sobj_data = sobj_data.copy()
                fields = sobj_data.pop("fields")
                sobj_data["last_modified_date"] = last_modified
                create_row(sess, SObject, sobj_data)
                for field in fields:
                    field["sobject"] = sobj_data["name"]
                    create_row(sess, Field, field)

            self.save_version(sess)

        engine.execute("vacuum")

    FormatVersion = "FormatVersion"
    CurrentFormatVersion = 2

    @property
    def version(self) -> int:
        rows = self.session.query(FileMetadata.value)
        rows = rows.filter(FileMetadata.name == self.FormatVersion)
        first_row = rows.one()
        assert first_row
        return int(first_row[0])

    def save_version(self, sess: "BufferedSession"):
        create_row(
            sess,
            FileMetadata,
            {"name": self.FormatVersion, "value": self.CurrentFormatVersion},
        )
        self.session.commit()


def create_row(buffered_session: "BufferedSession", model, valuesdict: dict):
    buffered_session.write_single_row(model.__tablename__, valuesdict)


class SilentMigration(Exception):
    pass


class BufferedSession:
    """Buffer writes to a SQL DB for faster performmance"""

    def __init__(self, engine: Engine, metadata: MetaData, max_buffer_size: int = 1000):
        self.buffered_rows = defaultdict(list)
        self.columns = {}
        self.engine = engine
        self.metadata = metadata
        self._prepare()
        self.max_buffer_size = max_buffer_size

    def __enter__(self, *args):
        self.session = create_session(bind=self.engine, autocommit=False)
        return self

    def __exit__(self, *args):
        self.close()

    def _prepare(self):
        # Setup table info used by the write-buffering infrastructure
        assert self.metadata.tables
        self.insert_statements = {}
        for tablename, model in self.metadata.tables.items():
            self.insert_statements[tablename] = model.insert(bind=self.engine)
            self.columns[tablename] = {
                colname: None for colname in model.columns.keys()
            }

    def write_single_row(self, tablename: str, row: Dict) -> None:
        # but first, normalize it so all keys have a value. SQLite Requires it.
        normalized_row = {}
        for key, default in self.columns[tablename].items():
            normalized_row[key] = row.get(key, default)

        # cache the value for later insert
        self.buffered_rows[tablename].append(normalized_row)

        # flush if buffer is full
        if len(self.buffered_rows[tablename]) > self.max_buffer_size:
            self.flush()

    def flush(self):
        for tablename, insert_statement in self.insert_statements.items():
            # seems much more efficient than line-by-line inserts.
            if self.buffered_rows[tablename]:
                self.session.execute(insert_statement, self.buffered_rows[tablename])
                self.buffered_rows[tablename] = []
        self.session.flush()

    def commit(self):
        self.flush()
        self.session.commit()

    def close(self) -> None:
        self.commit()
        self.session.close()


@contextmanager
def get_org_schema(
    sf,
    org_config,
    *,
    include_counts: bool = False,
    filters: T.Sequence[Filters] = (),
    patterns_to_ignore: T.Tuple[str, ...] = (),
    included_objects: T.List[str] = (),
    force_recache=False,
    logger=None,
):
    """
    Get a read-only representation of an org's schema.
    sf - simple_saleforce object
    org_config - an OrgConfig for the relevant org

    include_counts: query each queryable/retrievable object count.
                    This takes time and may even timeout in huge orgs!
    filters: A sequence of Filters which are the same as Salesforce SObject properties
            like .createable, .deletable etc. Objects that do not match are ignored.
            Two special filters exist:
                * Filters.extractable, which uses heuristics to limit to objects
                    that extract properly
                * Filters.populated, which limits to objects that have data in them.
                    This depends upon `include_counts`
    included_objects: Ignore objects not in this list. Stacks with other filters
    patterns_to_ignore: Strings in SQL %LIKE% syntax that match SObjects to be
                        ignored.

    force_recache: True - replace cache. False (default) - use/update cache is available.
    logger - replace the standard logger "cumulusci.salesforce_api.org_schema"

    This function take 5-20 seconds (or even more) depending on
    the complexity of the org. Call it at most once per task!
    """
    assert not isinstance(patterns_to_ignore, str)

    filters = set(filters)
    with org_config.get_orginfo_cache_dir(Schema.__module__) as directory:
        directory.mkdir(exist_ok=True, parents=True)
        schema_path = directory / "org_schema.db.gz"

        if force_recache and schema_path.exists():
            schema_path.unlink()

        if Filters.populated in filters:
            filters.add(Filters.queryable)
            filters.add(Filters.retrieveable)
            # experiment with removing this limitation by using limit 1 query instead
            assert include_counts, "Filters.populated depends on include_counts"
            patterns_to_ignore += NOT_COUNTABLE

        if Filters.extractable in filters:
            filters.add(Filters.queryable)
            filters.add(Filters.retrieveable)
            filters.add(Filters.createable)  # so we can load again later
            patterns_to_ignore += NOT_EXTRACTABLE

        logger = logger or getLogger(__name__)

        with ZippableTempDb() as tempdb, ExitStack() as closer:
            schema = None
            engine = tempdb.create_engine()
            if schema_path.exists():
                try:
                    cleanups_on_failure = []
                    tempdb.unzip_database(schema_path)
                    cleanups_on_failure.extend([schema_path.unlink, tempdb.clear])
                    schema = Schema(engine, schema_path, filters)

                    cleanups_on_failure.append(schema.close)
                    closer.callback(schema.close)
                    if schema.version != schema.CurrentFormatVersion:
                        raise SilentMigration(
                            "was created with older CumulusCI version"
                        )
                    assert schema.sobjects.first().name
                    schema.from_cache = True
                except Exception as e:
                    if not isinstance(e, SilentMigration):
                        logger.warning(
                            f"Cannot read `{schema_path}`. Recreating it. Reason `{e}`."
                        )
                    schema = None
                    for cleanup_action in reversed(cleanups_on_failure):
                        cleanup_action()

            if schema is None:
                Base.metadata.bind = engine
                Base.metadata.create_all()
                schema = Schema(engine, schema_path, filters)
                closer.callback(schema.close)
                schema.from_cache = False

            populated_objs = schema.populate_cache(
                sf,
                included_objects,
                filters,
                patterns_to_ignore,
                logger,
                include_counts=include_counts,
            )

            if Filters.populated in filters:
                # another way to compute this might be by querying the first ID
                objs_to_include = [
                    objname for objname, count in populated_objs.items() if count > 0
                ]
            else:
                objs_to_include = [objname for objname, _ in populated_objs.items()]

            schema.included_objects = objs_to_include
            schema.block_writing()
            # save a gzipped copy for later
            tempdb.zip_database(schema_path)
            yield schema


class ZippableTempDb:
    """A database that loads and saves from a tempdir to a zippped cache"""

    def __enter__(self) -> "ZippableTempDb":
        self.tempdir = TemporaryDirectory()
        self.tempfile = Path(self.tempdir.name) / "temp_org_schema.db"
        return self

    def __exit__(self, *args, **kwargs):
        self.clear()
        self.tempdir.cleanup()

    def zip_database(self, target_path: T.Union[FSResource, Path]):
        "Save a gzipped copy for later"
        zip_database(self.tempfile, target_path)

    def unzip_database(self, zipped_db: T.Union[FSResource, Path]):
        unzip_database(zipped_db, self.tempfile)

    def clear(self):
        if self.tempfile.exists():
            self.tempfile.unlink()

    def create_engine(self):
        return create_engine(f"sqlite:///{str(self.tempfile)}")


def populate_counts(sf, schema, objs_cached, logger) -> T.Dict[str, int]:
    objects_to_count = [objname for objname in objs_cached]
    counts, transports_errors, salesforce_errors = count_sobjects(sf, objects_to_count)
    errors = transports_errors + salesforce_errors
    for error in errors[0:10]:
        logger.warning(f"Error counting SObjects: {error}")

    if len(errors) > 10:
        logger.warning(f"{len(errors)} more counting errors suppressed")

    schema.add_counts(counts)
    schema.session.flush()
    return counts


class DescribeResponse(NamedTuple):
    """Result of a describe call from Salesforce"""

    status: int
    body: dict
    last_modified_date: str = None


class DescribeUpdate(NamedTuple):
    body: dict
    last_modified_date: str = None


lm_date = Optional[str]


def deep_describe(sf, objs: Dict[str, lm_date], logger) -> Iterable[DescribeUpdate]:
    """Fetch describe data for changed sobjects

    Fetch describe data for sobjects from the list 'objs'
    which have changed since last_modified_date (in HTTP
    proto format) and yield each object as a DescribeResponse object."""

    logger = logger or getLogger(__name__)
    with CompositeParallelSalesforce(sf, max_workers=8) as cpsf:
        responses, errors = cpsf.do_composite_requests(
            (
                {
                    "method": "GET",
                    "url": f"/services/data/v{sf.sf_version}/sobjects/{obj}/describe",
                    "referenceId": f"ref{obj}",
                    "httpHeaders": {"If-Modified-Since": last_modified_date or y2k},
                }
                for obj, last_modified_date in objs.items()
            )
        )

        for error in errors[0:5]:
            logger.warning(f"Error calling Salesforce API: {error}")
        if len(errors) > 5:
            logger.warning(f"...Total {len(errors)} errors")

        unrecoverable_errors = (
            error
            for error, matching_request in errors
            if not isinstance(error, RECOVERABLE_ERRORS)
        )
        first_unrecoverable_error = next(unrecoverable_errors, None)
        if first_unrecoverable_error:
            raise first_unrecoverable_error

        responses = (
            DescribeResponse(
                response["httpStatusCode"],
                response["body"],
                response["httpHeaders"].get("Last-Modified"),
            )
            for response in responses
        )

        changes = [
            DescribeUpdate(resp.body, resp.last_modified_date)
            for resp in responses
            if resp.status == 200
        ]
        unexpected = [resp for resp in responses if resp.status not in (200, 304)]
        for unknown in unexpected:  # pragma: no cover
            logger.warning(
                f"Unexpected describe reply. An SObject may be missing: {unknown}"
            )

        yield from changes


def ignore_based_on_properties(obj: dict, filters: T.Sequence[Filters]):
    return not all(obj.get(filter.name, True) for filter in filters)



""" a task for waiting on a specific custom settings value """

from simple_salesforce.exceptions import SalesforceError

from cumulusci.core.exceptions import TaskOptionsError
from cumulusci.core.utils import process_bool_arg, determine_managed_mode
from cumulusci.tasks.salesforce import BaseSalesforceApiTask


class CustomSettingValueWait(BaseSalesforceApiTask):
    """CustomSettingValueWait polls an org until the specific value exists in a custom settings field"""

    name = "CustomSettingValueWait"

    task_options = {
        "object": {
            "description": "Name of the Hierarchical Custom Settings object to query. Can include the %%%NAMESPACE%%% token. ",
            "required": True,
        },
        "field": {
            "description": "Name of the field on the Custom Settings to query. Can include the %%%NAMESPACE%%% token. ",
            "required": True,
        },
        "value": {
            "description": "Value of the field to wait for (String, Integer or Boolean). ",
            "required": True,
        },
        "managed": {
            "description": (
                "If True, will insert the project's namespace prefix.  "
                "Defaults to False or no namespace."
            ),
            "required": False,
        },
        "namespaced": {
            "description": (
                "If True, the %%%NAMESPACE%%% token "
                "will get replaced with the namespace prefix for the object and field."
                "Defaults to False."
            ),
            "required": False,
        },
        "poll_interval": {
            "description": (
                "Seconds to wait before polling for batch job completion. "
                "Defaults to 10 seconds."
            )
        },
    }

    def _run_task(self):
        self.poll_interval_s = int(self.options.get("poll_interval", 10))

        # Retrieve polling object/field/value options
        self.object_name = self.options["object"]
        self.field_name = self.options["field"]
        self.check_value = self.options["value"]

        # Process namespace tokens
        self._apply_namespace()

        # will block until poll_complete
        self._poll()

        self.logger.info("Value Matched.")

        return True

    def _poll_action(self):
        try:
            query_results = self.sf.query(self._object_query)
        except SalesforceError as e:
            message = e.content[0]["message"]
            if "No such column 'SetupOwnerId'" in message:
                message = "Only Hierarchical Custom Settings objects are supported."
            raise TaskOptionsError(f"Query Error: {message}")

        self.record = None
        for row in query_results["records"]:
            if row["SetupOwnerId"].startswith("00D"):
                self.record = row

        if self.record:
            self.poll_complete = not self._poll_again()
        else:
            self.logger.info(
                f"{self.field_name}: Looking for {self.check_value} and found no custom settings record"
            )

    def _poll_again(self):
        return not self.success

    def _apply_namespace(self):
        # Process namespace tokens
        namespace = self.project_config.project__package__namespace
        managed = determine_managed_mode(
            self.options, self.project_config, self.org_config
        )
        if "namespaced" in self.options:
            namespaced = process_bool_arg(self.options["namespaced"])
        else:
            namespaced = bool(namespace) and self.org_config.namespace == namespace

        namespace_prefix = ""
        if namespace and (managed or namespaced):
            namespace_prefix = namespace + "__"
        self.object_name = self.object_name.replace("%%%NAMESPACE%%%", namespace_prefix)
        self.field_name = self.field_name.replace("%%%NAMESPACE%%%", namespace_prefix)

    @property
    def success(self):
        lower_case_record = {k.lower(): v for k, v in self.record.items()}
        self.field_value = lower_case_record[self.field_name.lower()]

        if isinstance(self.field_value, bool):
            self.check_value = process_bool_arg(self.check_value)
            self.field_value = process_bool_arg(self.field_value)
        elif isinstance(self.field_value, (int, float)):
            self.check_value = float(self.check_value)
            self.field_value = float(self.field_value)
        elif isinstance(self.field_value, str):
            self.check_value = str(self.check_value).lower()
            self.field_value = str(self.field_value).lower()

        self.logger.info(
            f"{self.field_name}: Looking for {self.check_value} and found {self.field_value}"
        )
        return self.field_value == self.check_value

    @property
    def _object_query(self):
        return f"SELECT SetupOwnerId, {self.field_name} FROM {self.object_name}"
